-- Released under GPL v3
--------------------------------------------------------------
include("FLuaVector");
local toW = DirectionTypes.DIRECTION_WEST;
local toE = DirectionTypes.DIRECTION_EAST;
local toNW = DirectionTypes.DIRECTION_NORTHWEST;
local toNE = DirectionTypes.DIRECTION_NORTHEAST;
local toSW = DirectionTypes.DIRECTION_SOUTHWEST;
local toSE = DirectionTypes.DIRECTION_SOUTHEAST;
local NORIVER = FlowDirectionTypes.NO_FLOWDIRECTION;
local PLOT_MOUNTAIN = PlotTypes.PLOT_MOUNTAIN;
local PLOT_OCEAN = PlotTypes.PLOT_OCEAN;
local PLOT_HILLS = PlotTypes.PLOT_HILLS;
local PLOT_LAND = PlotTypes.PLOT_LAND;
local TERRAIN_OCEAN = TerrainTypes.TERRAIN_OCEAN;
local TERRAIN_COAST = TerrainTypes.TERRAIN_COAST;
local TERRAIN_GRASS = TerrainTypes.TERRAIN_COAST;

local wrapX = Map.IsWrapX();
local wrapY = Map.IsWrapY();
local w, h = Map.GetGridSize();

local validPlotTypes = {};
local terrainsData = {};
local featuresData = {};
local resourcesData = {};
local improvementsData = {};
local impassableTerrains = {};
local isRunningTests = false

local RsdArea = {};
local RsdPlot = {};

CivMap = Map;
RsdMap = 
{
	startingPlots = {},
	inlinePlots = {},
	areas = {},
};

local function sanitize(value)
	if sanitize == nil then return "<nil>" end
	return tostring(value)
end


--===============================================================================================
-- PLAYER
--===============================================================================================
RsdMap.GetStartingPlot = function(player) return RsdMap.startingPlots[player:GetID()] end
RsdMap.SetStartingPlot = function(player, plot) RsdMap.startingPlots[player:GetID()] = plot end


--===============================================================================================
-- PLOT
--===============================================================================================

RsdPlot.GetX = function(self) return self.x; end
RsdPlot.GetY = function(self) return self.y; end
RsdPlot.GetPlotType = function(self) return self.plotType; end
RsdPlot.GetTerrainType = function(self) return self.terrainType end
RsdPlot.GetFeatureType = function(self) return self.featureType end
RsdPlot.GetResourceType = function(self) return self.resourceType end
RsdPlot.GetNumResource = function(self) return self.numResource end
RsdPlot.GetImprovementType = function(self) return self.improvementType end
RsdPlot.GetContinentArtType = function(self) return self.art end
RsdPlot.GetArea = function(self) return self.area end
RsdPlot.IsWOfRiver = function(self) return self.riverW ~= NORIVER end
RsdPlot.IsNWOfRiver = function(self) return self.riverNW ~= NORIVER end
RsdPlot.IsNEOfRiver = function(self) return self.riverNE ~= NORIVER end
RsdPlot.GetRiverEFlowDirection = function(self) return self.riverW end
RsdPlot.GetRiverSEFlowDirection = function(self) return self.riverNW end
RsdPlot.GetRiverSWFlowDirection = function(self) return self.riverNE end
RsdPlot.IsRevealed = function(self) return self.revealed end

-- Not supported
RsdPlot.IsTradeRoute = function(self) return false end
RsdPlot.IsImprovementPillaged = function(self) return false end
RsdPlot.GetRevealedImprovementType = function(self) return self.improvementType end
RsdPlot.GetRevealedRouteType = function(self) return -1 end
RsdPlot.GetRevealedOwner = function(self) return -1 end
RsdPlot.IsCity = function(self) return false end

-- Setters
RsdPlot.SetContinentArtType = function(self, value) self.art = value or 0; end
RsdPlot.SetRevealed = function(self, value) self.revealed = value; end


RsdPlot.SetPlotType = function(self, value, recalculate) 
	-- Sanitize
	if (not value) or ((value ~= -1) and (not validPlotTypes[value])) then
		print("RSD_ERROR: wrong plot type: " .. sanitize(value))
		value = 0
	end

	if self.plotType == value then return end
	if recalculate == nil then recalculate = true end
	local isWaterTerrain = (self.terrainType == TERRAIN_COAST) or (self.terrainType == TERRAIN_OCEAN)
	local wasWater = self.plotType == PLOT_OCEAN
	local isWater = value == PLOT_OCEAN
	self.plotType = value


	-- Land plot with water terrain and recriprocally?
	-- Note: an ocean tile cannot be right next to a land tile: there must be a coast tile between both.
	if (self.terrainType == -1) or (isWaterTerrain ~= isWater) then
		if isWater then
			if self:IsAdjacentToLand() then
				self.terrainType = TERRAIN_COAST
			else
				self.terrainType = TERRAIN_OCEAN
			end
		else
			self.terrainType = TERRAIN_GRASS
		end
	end


	if wasWater ~= isWater and recalculate then
		-- Note: an ocean tile cannot be right next to a land tile: there must be a coast tile between both.
		for _, neighbor in pairs(self.neighbors) do
			if neighbor and neighbor.plotType == PLOT_OCEAN then
				if neighbor:IsAdjacentToLand() then
					neighbor:SetTerrainType(TERRAIN_COAST, recalculate)
				else
					neighbor:SetTerrainType(TERRAIN_OCEAN, recalculate)
				end
			end
		end

		-- Do we need to update the area?
		if self.area == -1 then return end

		-- Find unique matching area among neighbors
		local matchingArea = nil
		local recalculateAreas = false

		for _, neighbor in pairs(self.neighbors) do
			if neighbor then
				local neighborArea = neighbor:Area()
				if neighborArea and (isWater == neighborArea.isOcean) then
					if not matchingArea then
						matchingArea = neighborArea

					-- More than one matching area?
					elseif matchingArea ~= neighborArea then
						recalculateAreas = true
						break
					end
				end
			end
		end

		-- More than two neighbor areas? SHould not be possible
		if not recalculateAreas then
			local neighbor = self.neighbors[DirectionTypes.DIRECTION_NORTHWEST]
			local previousArea = neighbor and neighbor:Area()

			-- Count distinct neighboring areas
			local areaCount = 0;
			for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1 do
				local neighbor = self.neighbors[direction]
				local neighborArea = neighbor and neighbor:Area()

				if previousArea ~= neighborArea then areaCount = areaCount + 1 end
				previousArea = neighborArea;
			end

			recalculateAreas = areaCount > 2;
		end

		-- Set new area
		if recalculateAreas then
			RsdMap.RecalculateAreas()
		elseif matchingArea then
			self:SetArea(-1)
			self:SetArea(matchingArea:GetID())
		else
			self:SetArea(-1)
			RsdMap.CalculateAreas()
		end
	end
end 


RsdPlot.SetTerrainType = function(self, value, recalculate)
	-- Sanitize
	if (not value) or ((value ~= -1) and (not terrainsData[value])) then
		print("RSD_ERROR: wrong terrain type: " .. sanitize(value))
		value = 0
	end

	if self.terrainType == value then return end
	local isWaterTerrain = (value == TERRAIN_COAST) or (value == TERRAIN_OCEAN)
	local isWaterPlot = self.plotType == PLOT_OCEAN
	self.terrainType = value

	-- Set plot
	if isWaterPlot ~= isWaterTerrain then
		if isWaterTerrain then
			self:SetPlotType(PLOT_OCEAN, recalculate)
		else
			self:SetPlotType(PLOT_LAND, recalculate)
		end
	end
end

RsdPlot.SetFeatureType = function(self, value) 
	-- Sanitize
	if (not value) or ((value ~= -1) and (not featuresData[value])) then
		print("RSD_ERROR: wrong feature type: " .. sanitize(value))
		value = -1
	end

	self.featureType = value
end

RsdPlot.SetResourceType = function(self, value, count)
	if count == nil then 
		self:SetNumResource(0)
	else
		self:SetNumResource(count)
	end

	-- Sanitize
	if (not value) or ((value ~= -1) and (not resourcesData[value])) then
		print("RSD_ERROR: wrong resource type: " .. sanitize(value))
		value = -1
	end

	self.resourceType = value
end

RsdPlot.SetNumResource = function(self, value)
	-- Sanitize
	if (not value) or (type(value) ~= "number") or (value < 0) then
		print("RSD_ERROR: wrong resource qty: " .. sanitize(value))
		value = 0
	end
	self.numResource = value
end

RsdPlot.SetImprovementType = function(self, value) 
	-- Sanitize
	if (not value) or ((value ~= -1) and (not improvementsData[value])) then
		print("RSD_ERROR: wrong improvement type: " .. sanitize(value))
		value = -1
	end

	self.improvementType = value
end


-------------------------------------------------------------------------------------------------
RsdPlot.CalculateYield = function(self, value, bDisplay)
	local terrain = terrainsData[self.terrainType]
	local feature = featuresData[self.featureType]
	local resource = resourcesData[self.resourceType]
	local isHills = (self.plotType == PLOT_HILLS)
	local isRiver = self:IsRiverSide()
	local sum = 0
	-- No need to implement improvements

	if terrain then
		sum = sum + (terrain.yields[value] or 0)
		if isRiver then
			sum = sum + (terrain.riverYields[value] or 0)
		end
		if isHills then
			sum = sum + (terrain.hillsYields[value] or 0)
		end
	end

	if feature then
		sum = sum + (feature.yields[value] or 0)
		if isRiver then
			sum = sum + (feature.riverYields[value] or 0)
		end
		if isHills then
			sum = sum + (feature.hillsYields[value] or 0)
		end
	end
	
	if resource then
		sum = sum + (resource.yields[value] or 0)
	end
	return sum
end

-------------------------------------------------------------------------------------------------
RsdPlot.SetWOfRiver = function(self, value, flow) 
	if value then
		self.riverW = flow
	else
		self.riverW = NORIVER
	end
end

-------------------------------------------------------------------------------------------------
RsdPlot.SetNWOfRiver = function(self, value, flow) 
	if value then
		self.riverNW = flow
	else
		self.riverNW = NORIVER
	end
end

-------------------------------------------------------------------------------------------------
RsdPlot.SetNEOfRiver = function(self, value, flow) 
	if value then
		self.riverNE = flow
	else
		self.riverNE = NORIVER
	end
end

-------------------------------------------------------------------------------------------------
RsdPlot.GetLatitude = function(self) 
	local ratio;
	if wrapX or (not wrapY) then
		ratio = math.floor(100 * self.y / h);
	else
		ratio = math.floor(100 * self.x / w);
	end

	return math.abs(math.floor((180 * ratio - 9000) / 100))
end

-------------------------------------------------------------------------------------------------
RsdPlot.Area = function(self)
	return RsdMap.GetArea(self.area);
end

-------------------------------------------------------------------------------------------------
RsdPlot.SetArea = function(self, value)
	if self.area >= 0 then 
		RsdMap.GetArea(self.area):remove(self);
	end
	self.area = value; 
	if self.area >= 0 then 
		RsdMap.GetArea(self.area):add(self);
	end
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsRiverSide = function(self) 
	if self.riverNE ~= NORIVER then return true end
	if self.riverNW ~= NORIVER then return true end
	if self.riverW ~= NORIVER then return true end

	local nwNeighbor = self.neighbors[toNW];
	if nwNeighbor and nwNeighbor.riverNW ~= NORIVER then return true end

	local neNeighbor = self.neighbors[toNE];
	if neNeighbor and neNeighbor.riverNE ~= NORIVER then return true end

	local wNeighbor = self.neighbors[toW];
	if wNeighbor and wNeighbor.riverW ~= NORIVER then return true end

	return false
end
RsdPlot.IsRiver = RsdPlot.IsRiverSide

-------------------------------------------------------------------------------------------------
RsdPlot.IsCoastalLand = function(self) 
	if self.plotType ~= PLOT_OCEAN then
		for _, v in pairs(self.neighbors) do
			if v and (v.plotType == PLOT_OCEAN) then return true end
		end
	end
	return false;
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsCoastalLandNoLake = function(self) 
	if self.plotType ~= PLOT_OCEAN then
		for _, v in pairs(self.neighbors) do
			if v and (v.plotType == PLOT_OCEAN) and not v:IsLake() then return true end
		end
	end
	return false;
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsLake = function(self) 
	local area = self:Area();
	if not area then return false end
	return area:IsLake()
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsWater = function(self) 
	return self.plotType == PLOT_OCEAN
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsFreshWater = function(self) 
	if self:IsWater() then return false end
	if self:IsImpassable() then return false end
	if self.plotType == PLOT_MOUNTAIN then return false end
	if self:IsRiverSide() then return true end

	local feature = featuresData[self.featureType]
	if feature and feature.addsFreshWater then return true end

	for _, v in pairs(self.neighbors) do
		if v then
			if v:IsLake() then return true end
			local feature = featuresData[v.featureType]
			if feature and feature.addsFreshWater then return true end
		end
	end
	return false
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsAdjacentToShallowWater = function(self) 
	for _, v in pairs(self.neighbors) do
		if v and v.plotType == PLOT_OCEAN and v.terrainType == TERRAIN_COAST then return true end
	end
	return false
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsAdjacentToLand = function(self) 
	for _, v in pairs(self.neighbors) do
		if v and v.plotType ~= PLOT_OCEAN then return true end
	end
	return false
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsAdjacentToArea = function(self, area) 
	local areaID = area:GetID()
	for _, v in pairs(self.neighbors) do
		if v and v.area == areaID then return true end
	end
	return false
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsFlatlands = function(self) 
	return (self.plotType == PLOT_LAND)
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsHills = function(self) 
	return (self.plotType == PLOT_HILLS)
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsMountain = function(self) 
	return (self.plotType == PLOT_MOUNTAIN)
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsImpassable = function(self) 
	if impassableTerrains[self.terrainType] then return true end

	local feature = featuresData[self.featureType]
	if not feature then return false end
	if feature.impassable then return true end
	return false
end

-------------------------------------------------------------------------------------------------
RsdPlot.CanHaveFeature = function(self, featureID) 
	if self.featureType >= 0 then return false end
	if self.plotType == PLOT_MOUNTAIN then return false end

	local data = featuresData[featureID];
	if not data then return false end

	if not data.terrains[self.terrainType] then return false end
	if data.noCoast and self:IsCoastalLandNoLake() then return false end
	if data.noRiver and self:IsRiverSide() then return false end
	if data.requiresRiver and not self:IsRiverSide() then return false end
	if data.requiresFlatLands and self.plotType ~= PLOT_LAND then return false end
	if data.noAdjacent then 
		for _, plot in pairs(self.neighbors) do
			if plot and plot.featureType == featureID then return false end
		end
	end

	return true
end

-------------------------------------------------------------------------------------------------
RsdPlot.CanHaveResource = function(self, resourceID, ignoreLatitude) 
	if self.resourceType >= 0 then return false end
	if self.plotType == PLOT_MOUNTAIN then return false end
	if self:IsImpassable() then return false end

	local data = resourcesData[resourceID];
	if not data then return false end

	-- It must be on terrain A
	if self.featureType >= 0 then
		if not data.features[self.featureType] then return false end
		if not data.terrainsWithFeatures[self.terrainType] then return false end
	else
		if not data.terrains[self.terrainType] then return false end
	end

	-- Additional sea resources check
	if self.terrainType == TERRAIN_COAST then
		local dismiss = true
		for _, plot in pairs(self.neighbors) do
			if plot.plotType ~= PLOT_OCEAN then	
				dismiss = false
				break
			end
		end
		if dismiss then return false end
	end

	if not ignoreLatitude then
		local latitude = self:GetLatitude()
		if data.minLatitude and latitude < data.minLatitude then return false end
		if data.maxLatitude and latitude > data.maxLatitude then return false end
	end

	if self:Area():GetNumTiles() < data.minAreaSize then return false end
	if data.noRiverSide and self:IsRiverSide() then return false end

	if self.plotType == PLOT_HILLS then return data.hills end
	if self.plotType == PLOT_LAND then return data.flatLands end
	return true
end

-------------------------------------------------------------------------------------------------
RsdPlot.CanHaveImprovement = function(self, improvementID) 
	if self.plotType == PLOT_MOUNTAIN then return false end
	if self:IsImpassable() then return false end

	local data = improvementsData[improvementID]
	if not data then return false end

	if data.water ~= self:IsWater() then return false end

	local resource = resourcesData[self.resourceType]
	if self.resourceType >= 0 then
		if resource and resource.revealed and not data.resources[self.resourceType] and not data.buildableOnResources then return false end
	end

	local featureData = featuresData[self.featureType]
	if featureData and featureData.noImprovement then
		return false
	end

	if data.noFreshWater and self:IsFreshWater() then return false end
	if data.requiresFlatLands and self.plotType ~= PLOT_LAND then return false end
	if data.requiresFlatlandsOrFreshWater and self.plotType ~= PLOT_LAND and not self:IsFreshWater() then return false end
	if data.requiresFeature and not data.features[self.featureType] then return false end
	if data.coastal and not self:IsCoastalLand() then return false end

	if data.hillsMakesValid and self.plotType == PLOT_HILLS then return true end
	if data.freshWaterMakesValid and self:IsFreshWater() then return true end
	if data.riverSideMakesValid and self:IsRiverSide() then return true end
	if data.terrains[self.terrainType] then return true end
	if data.features[self.featureType] then return true end
	if data.resources[self.resourceType] and resource and resource.revealed then return true end

	return false
end

-------------------------------------------------------------------------------------------------
RsdPlot.IsStartingPlot = function(self) 
	for _, v in pairs(RsdMap.startingPlots) do
		if v == self then return true end
	end
	return false
end

-------------------------------------------------------------------------------------------------
-- We want to start a river at one of the vertices of the "self" hex. 
-- The vertex is randomly chosen. We must return one of the neighbors of this hex with that vertex on the SE.
RsdPlot.GetInlandCorner = function(self)
	local t = { 0, 1, 2 }
	Shuffle(t)

	for _, i in pairs(t) do
		local plot = self
		if i == 1 then 
			plot = self.neighbors[toW] 
		elseif i == 2 then 
			plot = self.neighbors[toNW] 
		end

		if plot and not plot:HasCoastAtSECorner() then
			return plot
		end
	end

	return nil
end

-------------------------------------------------------------------------------------------------
RsdPlot.HasCoastAtSECorner = function(self)
	local eastNeighbor = self.neighbors[toE];
	if eastNeighbor and eastNeighbor.plotType == PLOT_OCEAN then return true end

	local southEastNeighbor = self.neighbors[toSE];
	if southEastNeighbor and southEastNeighbor.plotType == PLOT_OCEAN then return true end
	
	return false
end

--===============================================================================================
-- AREA
--===============================================================================================
RsdArea.add = function(self, plot) 
	self.plots[plot] = true
	self.count = self.count + 1
end

-------------------------------------------------------------------------------------------------
RsdArea.remove = function(self, plot) 
	self.plots[plot] = false
	self.count = self.count - 1
end

-------------------------------------------------------------------------------------------------
RsdArea.contains = function(self, plot) 
	return self.plots[plot]
end

-------------------------------------------------------------------------------------------------
RsdArea.GetID = function(self) 
	return self._id
end

-------------------------------------------------------------------------------------------------
RsdArea.GetNumImprovements = function(self, improvementID) 
	if improvementID < 0 then return 0 end
	local count = 0;
	for plot, value in pairs(self.plots) do
		if value and plot:GetImprovementType() == improvementID then
			count = count + 1
		end
	end
	return count
end

-------------------------------------------------------------------------------------------------
RsdArea.GetNumRiverEdges = function(self) 
	local count = 0;
	for plot, value in pairs(self.plots) do
		if value then
			if plot.riverW ~= NORIVER then count = count + 1 end
			if plot.riverNW ~= NORIVER then count = count + 1 end
			if plot.riverNE ~= NORIVER then count = count + 1 end
		end
	end
	return count
end

-------------------------------------------------------------------------------------------------
RsdArea.GetNumTiles = function(self) 
	return self.count
end

-------------------------------------------------------------------------------------------------
RsdArea.IsLake = function(self)
	return self.isLake
end

-------------------------------------------------------------------------------------------------
RsdArea.IsWater = function(self)
	return self.isOcean
end



--===============================================================================================
-- MAP
--===============================================================================================
RsdMap.GetWorldSize = function()
	return CivMap.GetWorldSize()
end

-------------------------------------------------------------------------------------------------
RsdMap.GetSeaLevel = function()
	return CivMap.GetSeaLevel()
end

-------------------------------------------------------------------------------------------------
RsdMap.GetFractalFlags = function()
	return CivMap.GetFractalFlags()
end

-------------------------------------------------------------------------------------------------
RsdMap.GetClimate = function()
	return CivMap.GetClimate()
end

-------------------------------------------------------------------------------------------------
RsdMap.GetCustomOption = function(option)
	return CivMap.GetCustomOption(option)
end

-------------------------------------------------------------------------------------------------
RsdMap.IsWrapX = function()
	return wrapX
end

-------------------------------------------------------------------------------------------------
RsdMap.IsWrapY = function()
	return wrapY
end

-------------------------------------------------------------------------------------------------
RsdMap.GetGridSize = function() 
	return w, h 
end

-------------------------------------------------------------------------------------------------
RsdMap.GetNumPlots = function()
	return w * h
end

-------------------------------------------------------------------------------------------------
RsdMap.GetPlotByIndex = function(i)
	return RsdMap.inlinePlots[i + 1]
end

-------------------------------------------------------------------------------------------------
RsdMap.Rand = function(max, generator)
	return CivMap.Rand(max, generator)
end

-------------------------------------------------------------------------------------------------
RsdMap.GetRandomResourceQuantity = function(resourceID)
	return CivMap.GetRandomResourceQuantity(resourceID)
end

-------------------------------------------------------------------------------------------------
RsdMap.DefaultContinentStamper = function()
	RsdMap.useDefaultContinentStamper = true
	-- Do nothing: this is the last operation done by GenerateMap, so we will just invoke the real CivMap.DefaultContinentStamper
end

-------------------------------------------------------------------------------------------------
RsdMap.ChangeAIMapHint = function(value)
	RsdMap.mapHint = value
end

-------------------------------------------------------------------------------------------------
RsdMap.GetPlot = function(x, y)
	if wrapX then
		x = (x + w) % w
	end
	if wrapY then
		y = (y + h) % h
	end
	local index = y * w + x + 1
	return RsdMap.inlinePlots[index]
end

-------------------------------------------------------------------------------------------------
-- GetPlotXY(0, 1, -5, -1) -> (35, 0)
-- GetPlotXY(0, 2, -5, -1) -> (34, 1)
RsdMap.GetPlotXY = function(x, y, dx, dy)
	local x2 = x + dx + dy * 0.5 + (y % 2) * 0.5
	local y2 = y + dy
	x2 = math.floor(x2)

	return RsdMap.GetPlot(x2, y2), x2, y2
end

-------------------------------------------------------------------------------------------------
RsdMap.PlotDistance = function(x1, y1, x2, y2)
	return CivMap.PlotDistance(x1, y1, x2, y2)
	--[[local xc1 = x1
	local xc2 = x2
	if (y1 % 2) == 1 then xc1 = xc1 + 0.5 end
	if (y2 % 2) == 1 then xc2 = xc2 + 0.5 end
	local squareDist = (y2 - y1) * (y2 - y1) + (xc2 - xc1) * (xc2 - xc1)
	return math.sqrt(squareDist)]]
end

-------------------------------------------------------------------------------------------------
RsdMap.PlotXYWithRangeCheck = function(x, y, dx, dy, range)
	local plot, x2, y2 = RsdMap.GetPlotXY(x, y, dx, dy)
	local distance = CivMap.PlotDistance(x, y, x2, y2)
	--print("PlotXYWithRangeCheck", distance, range, tostring(plot))

	if distance > range then return nil end
	return plot
end

-------------------------------------------------------------------------------------------------
RsdMap.PlotDirection = function(x, y, direction)
	if direction == toNW then
		if (y % 2) == 1 then
			return RsdMap.GetPlot(x, y + 1)
		else
			return RsdMap.GetPlot(x - 1, y + 1)
		end
	elseif direction == toNE then
		if (y % 2) == 1 then
			return RsdMap.GetPlot(x + 1, y + 1)
		else
			return RsdMap.GetPlot(x, y + 1)
		end
	elseif direction == toSW then
		if (y % 2) == 1 then
			return RsdMap.GetPlot(x, y - 1)
		else
			return RsdMap.GetPlot(x - 1, y - 1)
		end
	elseif direction == toSE then
		if (y % 2) == 1 then
			return RsdMap.GetPlot(x + 1, y - 1)
		else
			return RsdMap.GetPlot(x, y - 1)
		end
	elseif direction == toW then
		return RsdMap.GetPlot(x - 1, y)
	elseif direction == toE then
		return RsdMap.GetPlot(x + 1, y)
	end
end

-------------------------------------------------------------------------------------------------
RsdMap.Areas = function()
	local iter = function(areas, area)
		if area == nil then 
			return select(2, next(areas, nil))
		else
			return select(2, next(areas, area:GetID()))
		end
	end

	return iter, RsdMap.areas, nil
end

-------------------------------------------------------------------------------------------------
RsdMap.GetArea = function(id)
	if id < 0 then return nil end
	if not RsdMap.areas[id] then
		RsdMap.areas[id] = { _id = id, count = 0, plots = {} }
		for k, v in pairs(RsdArea) do
			RsdMap.areas[id][k] = v
		end
	end
	return RsdMap.areas[id]
end

-------------------------------------------------------------------------------------------------
RsdMap.FindBiggestArea = function(ocean)
	local bestArea = nil
	local bestCount = 0
	for k, v in pairs(RsdMap.areas) do
		if (v.isLandmass and not ocean) or (v.isOcean and ocean) then 
			local count = v:GetNumTiles()
			--print(k, count, tostring(v.group))
			if count > bestCount then
				bestCount = count
				bestArea = v
			end
		end
	end
	return bestArea
end

-------------------------------------------------------------------------------------------------
RsdMap.FindWater = function(plot, range, isFreshWater)
	for dx = -range, range, 1 do
		for dy = -range, range, 1 do
			local otherPlot = Map.PlotXYWithRangeCheck(plot.x, plot.y, dx, dy, range)
			if otherPlot then
				if isFreshWater then
					if otherPlot:IsFreshWater() then 
						--print("found fresh water on", dx, dy, range)
						return true 
					end
				else
					if otherPlot:IsWater() then 
						--print("found water on", dx, dy, range)
						return true 
					end
				end
			end
		end
	end
	--[[if isFreshWater then
		print("fresh water not found", range)
	else
		print("water not found", range)
	end]]
	return false
end

-------------------------------------------------------------------------------------------------
RsdMap.GetLandPlots = function()
	local count = 0;
	for _, plot in ipairs(RsdMap.inlinePlots) do
		if plot.plotType ~= PLOT_OCEAN then count = count + 1 end
	end
	return count
end

-------------------------------------------------------------------------------------------------
RsdMap.GetNumResources = function(resourceID)
	if resourceID < 0 then return 0 end
	local count = 0;
	for _, plot in ipairs(RsdMap.inlinePlots) do
		if plot.resourceType == resourceID then count = count + 1 end
	end
	return count
end

-------------------------------------------------------------------------------------------------
RsdMap.GetNumResourcesOnLand = function(resourceID)
	if resourceID < 0 then return 0 end
	local count = 0;
	for _, plot in ipairs(RsdMap.inlinePlots) do
		if (plot.plotType ~= PLOT_OCEAN) and (plot.resourceType == resourceID) then 
			count = count + 1 
		end
	end
	return count
end

-------------------------------------------------------------------------------------------------
RsdMap.ValidateCoast = function()
	-- There must be a coast between an ocean tile and a non-sea tile.
	-- The game runs a similar pass after all map generation.
	for _, plot in ipairs(RsdMap.inlinePlots) do
		if plot.terrainType == TERRAIN_OCEAN then
			for _, neighbor in pairs(plot.neighbors) do
				if neighbor and neighbor.plotType ~= PLOT_OCEAN then
					plot.terrainType = TERRAIN_COAST
				end
			end
		end
	end
	return true
end

-------------------------------------------------------------------------------------------------
local areaGroups = {};
areaGroups[PLOT_OCEAN] = 1;
areaGroups[PLOT_HILLS] = 2;
areaGroups[PLOT_LAND] = 2;
areaGroups[PLOT_MOUNTAIN] = 2;

local function GetAreaGroup(plot)
	if (not areaGroups[plot.plotType]) then 
		error("no area group for: "..tostring(plot.plotType))
		return -1
	end
	if plot:IsImpassable() then 
		return areaGroups[plot.plotType] + 3
	else
		return areaGroups[plot.plotType]
	end
end

local function FillArea(sourcePlot, areaID, group)
	local area = RsdMap.GetArea(areaID)
	sourcePlot.area = areaID

	local toFill = { sourcePlot }
	local i = 1
	while i <= #toFill do
		local plot = toFill[i]
		for _, neighbor in pairs(plot.neighbors) do
			if neighbor and neighbor.area == -1 and GetAreaGroup(neighbor) == group then
				table.insert(toFill, neighbor)
				neighbor.area = areaID
			end
		end
		i = i + 1
	end

	for _, plot in ipairs(toFill) do
		area:add(plot)
	end
end

-------------------------------------------------------------------------------------------------
RsdMap.CalculateAreas = function()
	print("Calculate areas")

	local areaID = 0;
	for i in pairs(RsdMap.areas) do
		if (i + 1) > areaID then areaID = i + 1 end
	end

	-- Create new areas for unassigned plots
	for _, plot in ipairs(RsdMap.inlinePlots) do
		if plot.area < 0 then
			FillArea(plot, areaID, GetAreaGroup(plot));
			areaID = areaID + 1;
		end
	end	

	-- Guess groups
	for _, area in pairs(RsdMap.areas) do
		area.group = -1
		area.isOcean = false
		area.isLandmass = false
		for plot, value in pairs(area.plots) do
			if value then
				area.group = GetAreaGroup(plot) or -1;
				break
			end
		end
		area.isLandmass = area.group == 2
		area.isOcean = area.group == 1
		assert(area.group)
	end

	-- Flag lakes
	for id, area in pairs(RsdMap.areas) do
		if area:GetNumTiles() > GameDefines.LAKE_MAX_AREA_SIZE then
			area.isLake = false
		else
			area.isLake = true
			local count = 0

			-- Must be full of coasts and with a size lesser than LAKE_MAX_AREA_SIZE
			for plot, value in pairs(area.plots) do
				if value then 
					count = count + 1 
					if count > GameDefines.LAKE_MAX_AREA_SIZE then 
						area.isLake = false 
						break
					end
					if (plot.terrainType ~= TERRAIN_COAST) then 
						area.isLake = false 
						break
					end
				end
			end

			-- Must not be adjacent to another water area
			if area.isLake then
				--print("found candidate")
				for plot, value in pairs(area.plots) do
					if value then 
						for _, adjacentArea in pairs(RsdMap.areas) do
							if adjacentArea and area ~= adjacentArea and plot:IsAdjacentToArea(adjacentArea) then
								if (adjacentArea.group % 3) == 1 then 
									area.isLake = false
									break
								end
							end
						end
					end
					if not area.isLake then break end
				end
			end
		end
	end

	if not isRunningTests then coroutine.yield() end
end

-------------------------------------------------------------------------------------------------
RsdMap.RecalculateAreas = function()
	RsdMap.areas = {}
	for _, plot in ipairs(RsdMap.inlinePlots) do
		plot.area = -1
	end	

	RsdMap.CalculateAreas()
end

--===============================================================================================
-- ENABLE/DISABLE OVERRIDES
--===============================================================================================
local oldGetStartingPlot, oldSetStartingPlot
function EnableRsdMap()
	RsdMap.useDefaultContinentStamper = false
	RsdMap.mapHint = 0
	RsdMap.inlinePlots = {}
	RsdMap.areas = {}

	for i = 0, h - 1, 1 do
		for j = 0, w - 1, 1 do
			local plot = {
				x = j,
				y = i,
				plotType = PLOT_OCEAN,
				terrainType = -1,
				featureType = -1,
				resourceType = -1,
				improvementType = -1,
				NumResource = 0,
				revealed = false,
				area = -1,
				art = 0,
				riverNE = NORIVER,
				riverNW = NORIVER,
				riverW = NORIVER,
				neighbors = {},
			}

			for k, v in pairs(RsdPlot) do
				plot[k] = v
			end

			table.insert(RsdMap.inlinePlots, plot)
		end
	end

	for _, plot in ipairs(RsdMap.inlinePlots) do
		local x = plot:GetX()
		local y = plot:GetY()
		plot.neighbors[toNW] = RsdMap.PlotDirection(x, y, toNW);
		plot.neighbors[toNE] = RsdMap.PlotDirection(x, y, toNE);
		plot.neighbors[toSW] = RsdMap.PlotDirection(x, y, toSW);
		plot.neighbors[toSE] = RsdMap.PlotDirection(x, y, toSE);
		plot.neighbors[toW] = RsdMap.PlotDirection(x, y, toW);
		plot.neighbors[toE] = RsdMap.PlotDirection(x, y, toE);
	end

	for i = 0, GameDefines.MAX_CIV_PLAYERS - 1 do
		oldGetStartingPlot = Players[i].GetStartingPlot
		oldSetStartingPlot = Players[i].SetStartingPlot
		Players[i].SetStartingPlot = RsdMap.SetStartingPlot
		Players[i].GetStartingPlot = RsdMap.GetStartingPlot;
	end

	Map = RsdMap;
end

-------------------------------------------------------------------------------------------------
function DisableRsdMap()
	Map = CivMap

	for i = 0, GameDefines.MAX_CIV_PLAYERS - 1 do
		Players[i].SetStartingPlot = oldSetStartingPlot
		Players[i].GetStartingPlot = oldGetStartingPlot
	end
end

--===============================================================================================
-- INITIALIZATION
--===============================================================================================
local yields = {}
for row in GameInfo.Yields() do
	yields[row.Type] = row.ID
end

validPlotTypes[PLOT_MOUNTAIN] = true
validPlotTypes[PLOT_OCEAN] = true
validPlotTypes[PLOT_HILLS] = true
validPlotTypes[PLOT_LAND] = true

-------------------------------------------------------------------------------------------------
for row in GameInfo.Terrains() do
	if row.Impassable then
		impassableTerrains[row.ID] = true
	end

	terrainsData[row.ID] = { yields = {}, riverYields = {}, hillsYields = {} }

	for subRow in GameInfo.Terrain_Yields{TerrainType = row.Type} do
		local yieldID = yields[subRow.YieldType] or -1
		terrainsData[row.ID].yields[yieldID] = subRow.Yield
	end
	for subRow in GameInfo.Terrain_RiverYieldChanges{TerrainType = row.Type} do
		local yieldID = yields[subRow.YieldType] or -1
		terrainsData[row.ID].riverYields[yieldID] = subRow.Yield
	end
	for subRow in GameInfo.Terrain_HillsYieldChanges{TerrainType = row.Type} do
		local yieldID = yields[subRow.YieldType] or -1
		terrainsData[row.ID].hillsYields[yieldID] = subRow.Yield
	end
end

-------------------------------------------------------------------------------------------------
for row in GameInfo.Features() do
	featuresData[row.ID] = {
		type = row.Type,
		noCoast = row.NoCoast,
		noRiver = row.NoRiver,
		noAdjacent = row.NoAdjacent,
		noImprovement = row.NoImprovement,
		requiresFlatLands = row.RequiresFlatlands,
		requiresRiver = row.RequiresRiver,
		addsFreshWater = row.AddsFreshWater,
		impassable = row.Impassable,
		terrains = {},
		riverYields = {}, 
		hillsYields = {},
		yields = {}, 
	}

	for subRow in GameInfo.Feature_TerrainBooleans{FeatureType = row.Type} do
		local terrain = GameInfo.Terrains[subRow.TerrainType]
		if terrain then
			featuresData[row.ID].terrains[terrain.ID] = true
		end
	end
	for subRow in GameInfo.Feature_YieldChanges{FeatureType = row.Type} do
		local yieldID = yields[subRow.YieldType] or -1
		featuresData[row.ID].yields[yieldID] = subRow.Yield
	end
	for subRow in GameInfo.Feature_RiverYieldChanges{FeatureType = row.Type} do
		local yieldID = yields[subRow.YieldType] or -1
		featuresData[row.ID].riverYields[yieldID] = subRow.Yield
	end
	for subRow in GameInfo.Feature_HillsYieldChanges{FeatureType = row.Type} do
		local yieldID = yields[subRow.YieldType] or -1
		featuresData[row.ID].hillsYields[yieldID] = subRow.Yield
	end
end

-------------------------------------------------------------------------------------------------
for row in GameInfo.Resources() do
	resourcesData[row.ID] = {
		type = row.Type,
		minAreaSize = row.MinAreaSize,
		minLatitude = row.MinLatitude,
		maxLatitude = row.MaxLatitude,
		hills = row.Hills,
		flatLands = row.Flatlands,
		noRiverSide = row.NoRiverSide,
		terrainsWithFeatures = {},
		revealed = not row.TechReveal,
		features = {},
		terrains = {},
		yields = {},
	}

	for subRow in GameInfo.Resource_TerrainBooleans{ResourceType = row.Type} do
		local terrain = GameInfo.Terrains[subRow.TerrainType]
		if terrain then
			resourcesData[row.ID].terrains[terrain.ID] = true
		end
	end

	for subRow in GameInfo.Resource_FeatureTerrainBooleans{ResourceType = row.Type} do
		local terrain = GameInfo.Terrains[subRow.TerrainType]
		if terrain then
			resourcesData[row.ID].terrainsWithFeatures[terrain.ID] = true
		end
	end

	for subRow in GameInfo.Resource_FeatureBooleans{ResourceType = row.Type} do
		local feature = GameInfo.Features[subRow.FeatureType]
		if feature then
			resourcesData[row.ID].features[feature.ID] = true
		end
	end

	for subRow in GameInfo.Resource_YieldChanges{ResourceType = row.Type} do
		local yieldID = yields[subRow.YieldType] or -1
		resourcesData[row.ID].yields[yieldID] = subRow.Yield
	end
end

-------------------------------------------------------------------------------------------------
for row in GameInfo.Improvements() do
	improvementsData[row.ID] = {
		type = row.Type,
		hillsMakesValid = row.HillsMakesValid,
		freshWaterMakesValid = row.FreshWaterMakesValid,
		riverSideMakesValid = row.RiverSideMakesValid,
		noFreshWater = row.NoFreshWater,
		requiresFeature = row.RequiresFeature,
		requiresFlatLands = row.RequiresFlatlands,
		requiresFlatlandsOrFreshWater = row.RequiresFlatlandsOrFreshWater,
		buildableOnResources = row.BuildableOnResources,
		barbarianCamp = row.BarbarianCamp,
		coastal = row.Coastal,
		water = row.Water,
		resources = {},
		terrains = {},
		features = {},
	}

	for subRow in GameInfo.Improvement_ValidTerrains{ImprovementType = row.Type} do
		local terrain = GameInfo.Terrains[subRow.TerrainType]
		if terrain then
			improvementsData[row.ID].terrains[terrain.ID] = true
		end
	end

	for subRow in GameInfo.Improvement_ValidFeatures{ImprovementType = row.Type} do
		local feature = GameInfo.Features[subRow.FeatureType]
		if feature then
			improvementsData[row.ID].features[feature.ID] = true
		end
	end

	for subRow in GameInfo.Improvement_ResourceTypes{ImprovementType = row.Type} do
		local resource = GameInfo.Resources[subRow.ResourceType]
		if resource and subRow.ResourceMakesValid then
			improvementsData[row.ID].resources[resource.ID] = true
		end
	end

end


--===============================================================================================
-- TESTING
--===============================================================================================
local function AssertSamePlots(plot1, plot2)
	if plot1 then
		if not plot2 then
			error("plot1="..tostring(plot1)..", plot2=nil")
		end
		if (plot1:GetX() ~= plot2:GetX()) or (plot1:GetY() ~= plot2:GetY()) then
			error("different coords: "..plot1:GetX()..", "..plot1:GetY()..", "..plot2:GetX()..", "..plot2:GetY());
		end
	else
		if plot2 then
			error("plot1=nil, plot2="..tostring(plot2))
		end
	end
end

local function HighlightHex(plot)
	Events.SerialEventHexHighlight(ToHexFromGrid(Vector2(plot:GetX(), plot:GetY())), true, Vector4(1, 0, 0, 1));
end

local function RunTests()
	local directions = { toNE, toE, toSE, toNW, toW, toSW }
	isRunningTests = true

	Events.ClearHexHighlights();
	CivMap.RecalculateAreas();
	EnableRsdMap();
	
	print("Preparing tests");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);
		assert(rsdPlot);

		rsdPlot:SetPlotType(civPlot:GetPlotType())
		rsdPlot:SetTerrainType(civPlot:GetTerrainType())
		rsdPlot:SetFeatureType(civPlot:GetFeatureType())
		rsdPlot:SetResourceType(civPlot:GetResourceType())
		rsdPlot:SetImprovementType(civPlot:GetImprovementType())
		rsdPlot:SetContinentArtType(civPlot:GetContinentArtType())
		rsdPlot:SetNumResource(civPlot:GetNumResource())
		rsdPlot:SetWOfRiver(civPlot:IsWOfRiver(), civPlot:GetRiverEFlowDirection())
		rsdPlot:SetNWOfRiver(civPlot:IsNWOfRiver(), civPlot:GetRiverSEFlowDirection())
		rsdPlot:SetNEOfRiver(civPlot:IsNEOfRiver(), civPlot:GetRiverSWFlowDirection())
		rsdPlot:SetArea(civPlot:GetArea())
	end
	RsdMap.CalculateAreas();

	print("Testing basic getters/setters");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);
		--print(rsdPlot:GetY(), rsdPlot:GetLatitude(), civPlot:GetLatitude())

		assert(rsdPlot:GetPlotType() == civPlot:GetPlotType());
		assert(rsdPlot:GetTerrainType() == civPlot:GetTerrainType());
		assert(rsdPlot:GetFeatureType() == civPlot:GetFeatureType());
		assert(rsdPlot:GetNumResource() == civPlot:GetNumResource());
		assert(rsdPlot:GetResourceType() == civPlot:GetResourceType());
		assert(rsdPlot:GetImprovementType() == civPlot:GetImprovementType());
		assert(rsdPlot:GetContinentArtType() == civPlot:GetContinentArtType());
		assert(rsdPlot:GetRiverSWFlowDirection() == civPlot:GetRiverSWFlowDirection());
		assert(rsdPlot:GetRiverSEFlowDirection() == civPlot:GetRiverSEFlowDirection());
		assert(rsdPlot:GetRiverEFlowDirection() == civPlot:GetRiverEFlowDirection());
		assert(rsdPlot:GetLatitude() == civPlot:GetLatitude());
		assert(rsdPlot:IsNWOfRiver() == civPlot:IsNWOfRiver());
		assert(rsdPlot:IsNEOfRiver() == civPlot:IsNEOfRiver());
		assert(rsdPlot:IsWOfRiver() == civPlot:IsWOfRiver());
		assert(rsdPlot:GetArea() == civPlot:GetArea());
		assert(rsdPlot:GetX() == civPlot:GetX());
		assert(rsdPlot:GetY() == civPlot:GetY());
	end

	print("Testing GetPlot");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlot(civPlot:GetX(), civPlot:GetY());
		AssertSamePlots(civPlot, rsdPlot);
	end

	print("Testing PlotDirection");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);

		for _, direction in ipairs(directions) do
			local rsdNeighbor = RsdMap.PlotDirection(rsdPlot:GetX(), rsdPlot:GetY(), direction);
			local civNeighbor = CivMap.PlotDirection(civPlot:GetX(), civPlot:GetY(), direction);
			AssertSamePlots(rsdNeighbor, civNeighbor);
		end
	end

	print("Testing advanced getters");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);
		--print(civPlot:GetX(), civPlot:GetY())

		assert(rsdPlot:IsRiverSide() == civPlot:IsRiverSide());
		assert(rsdPlot:IsCoastalLand() == civPlot:IsCoastalLand());
		assert(rsdPlot:IsWater() == civPlot:IsWater());
		assert(rsdPlot:IsHills() == civPlot:IsHills());
		assert(rsdPlot:IsMountain() == civPlot:IsMountain());
		assert(rsdPlot:IsFlatlands() == civPlot:IsFlatlands());
		assert(rsdPlot:IsImpassable() == civPlot:IsImpassable());
		assert(rsdPlot:IsAdjacentToLand() == civPlot:IsAdjacentToLand());
	end

	print("Testing IsAdjacentToArea");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);

		for rsdArea in RsdMap.Areas() do
			local civArea = CivMap.GetArea(rsdArea:GetID());
			assert(rsdPlot:IsAdjacentToArea(rsdArea) == civPlot:IsAdjacentToArea(civArea));
		end
	end

	print("Testing area getters");
	for rsdArea in RsdMap.Areas() do
		local civArea = CivMap.GetArea(rsdArea:GetID());
		--print(rsdArea:GetID(), rsdArea:GetNumTiles(), civArea:GetNumTiles(), rsdArea:GetNumRiverEdges(), civArea:GetNumRiverEdges())
		assert(rsdArea:GetNumTiles() == civArea:GetNumTiles());
		assert(rsdArea:GetNumRiverEdges() == civArea:GetNumRiverEdges());
	end

	print("Testing IsLake");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);
		--print(civPlot:GetX(), civPlot:GetY(), civPlot:GetArea(), rsdPlot:GetArea(), civPlot:Area():GetNumTiles(), rsdPlot:Area():GetNumTiles(), civPlot:IsImpassable(), rsdPlot:IsImpassable())

		assert(rsdPlot:IsLake() == civPlot:IsLake());
	end

	print("Testing water getters");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);
		--print(civPlot:GetX(), civPlot:GetY(), civPlot:IsFreshWater(), rsdPlot:IsFreshWater(), civPlot:IsImpassable(), rsdPlot:IsImpassable())

		assert(rsdPlot:IsFreshWater() == civPlot:IsFreshWater());
		assert(rsdPlot:IsAdjacentToShallowWater() == civPlot:IsAdjacentToShallowWater());
	end

	--[[for y = 0, 4 do
		local civPlot = CivMap.GetPlot(0, y);
		local rsdPlot = RsdMap.GetPlot(0, y);
		for dy = -5, 5, 1 do
			local rsdOther = RsdMap.GetPlotXY(rsdPlot:GetX(), rsdPlot:GetY(), -5, dy)
			local civOther = CivMap.GetPlotXY(civPlot:GetX(), civPlot:GetY(), -5, dy)

			local x1, y1, x2, y2
			if rsdOther then
				x1, y1 = rsdOther:GetX(), rsdOther:GetY()
			else
				x1, y1 = "nil", "nil"
			end
			if civOther then
				x2, y2 = civOther:GetX(), civOther:GetY()
			else
				x2, y2 = "nil", "nil"
			end
			if x1 == x2 and y1 == y2 then
				print(y, dy, x1, y1)
			else
				print(y, dy, x2, y2, x1, y1, "bad")
			end
		end
	end]]

	print("Testing GetPlotXY & PlotXYWithRangeCheck");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);

		--print(rsdPlot:GetX(), rsdPlot:GetY());
		for range = 5, 0, -1 do 
			for dx = -5, 5, 1 do
				for dy = -5, 5, 1 do
					--print("dx = "..dx..", dy = "..dy)
					local rsdOther = RsdMap.GetPlotXY(rsdPlot:GetX(), rsdPlot:GetY(), dx, dy)
					local civOther = CivMap.GetPlotXY(civPlot:GetX(), civPlot:GetY(), dx, dy)
					AssertSamePlots(rsdOther, civOther);

					if civOther then
						local d1 = CivMap.PlotDistance(civPlot:GetX(), civPlot:GetY(), civOther:GetX(), civOther:GetY())
						local d2 = CivMap.PlotDistance(rsdPlot:GetX(), rsdPlot:GetY(), rsdOther:GetX(), rsdOther:GetY())
						assert(d1 == d2)
					end

					local rsdOtherRange = RsdMap.PlotXYWithRangeCheck(rsdPlot:GetX(), rsdPlot:GetY(), dx, dy, range)
					local civOtherRange = CivMap.PlotXYWithRangeCheck(civPlot:GetX(), civPlot:GetY(), dx, dy, range)

					--[[print(tostring(rsdOther), tostring(rsdOtherRange), tostring(civOther), tostring(civOtherRange))
					if rsdOtherRange and civOtherRange then
						print(rsdOtherRange:GetX(), rsdOtherRange:GetY(), civOtherRange:GetX(), civOtherRange:GetY());
					end]]

					AssertSamePlots(rsdOtherRange, civOtherRange);
				end
			end
		end
	end

	print("Testing FindWater");
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);

		for range = 5, 0, -1 do 
			--print(rsdPlot:GetX(), rsdPlot:GetY(), range, RsdMap.FindWater(rsdPlot, range, true), CivMap.FindWater(civPlot, range, true), RsdMap.FindWater(rsdPlot, range, false), CivMap.FindWater(civPlot, range, false));
			assert(RsdMap.FindWater(rsdPlot, range, true) == CivMap.FindWater(civPlot, range, true));
			assert(RsdMap.FindWater(rsdPlot, range, false) == CivMap.FindWater(civPlot, range, false));
		end
	end

	print("Testing CanHaveFeature");
	for id, v in pairs(featuresData) do
		for i = 0, Map.GetNumPlots()-1, 1 do
			local civPlot = CivMap.GetPlotByIndex(i);
			local rsdPlot = RsdMap.GetPlotByIndex(i);

			if rsdPlot:CanHaveFeature(id) ~= civPlot:CanHaveFeature(id) then
				print(v.type, civPlot:GetX(), civPlot:GetY(), rsdPlot:CanHaveFeature(id), civPlot:CanHaveFeature(id));
				HighlightHex(rsdPlot)
			end
			assert(rsdPlot:CanHaveFeature(id) == civPlot:CanHaveFeature(id));
		end
	end

	print("Testing CanHaveResource");
	for id, v in pairs(resourcesData) do
		for i = 0, Map.GetNumPlots()-1, 1 do
			local civPlot = CivMap.GetPlotByIndex(i);
			local rsdPlot = RsdMap.GetPlotByIndex(i);

			if rsdPlot:CanHaveResource(id) ~= civPlot:CanHaveResource(id) then
				print(civPlot:GetX(), civPlot:GetY(), v.type, rsdPlot:CanHaveResource(id), civPlot:CanHaveResource(id), rsdPlot:CanHaveResource(id, true), civPlot:CanHaveResource(id, true));
				HighlightHex(rsdPlot)
				break
			end

			assert(rsdPlot:CanHaveResource(id, false) == civPlot:CanHaveResource(id, false));
			assert(rsdPlot:CanHaveResource(id, true) == civPlot:CanHaveResource(id, true));
		end
	end

	print("Testing CanHaveImprovement");
	for id, v in pairs(improvementsData) do
		--print(v.type)
		for i = 0, Map.GetNumPlots()-1, 1 do
			local civPlot = CivMap.GetPlotByIndex(i);
			local rsdPlot = RsdMap.GetPlotByIndex(i);

			if rsdPlot:CanHaveImprovement(id) ~= civPlot:CanHaveImprovement(id) then
				print(v.type, civPlot:GetX(), civPlot:GetY(), rsdPlot:CanHaveImprovement(id), civPlot:CanHaveImprovement(id));
				HighlightHex(rsdPlot)
				break
			end

			assert(rsdPlot:CanHaveImprovement(id) == civPlot:CanHaveImprovement(id));
		end
	end

	print("Testing GetNumResources");
	for row in GameInfo.Resources() do
		assert(RsdMap.GetNumResources(row.ID) == CivMap.GetNumResources(row.ID))
		assert(RsdMap.GetNumResourcesOnLand(row.ID) == CivMap.GetNumResourcesOnLand(row.ID))
	end
	assert(RsdMap.GetNumResourcesOnLand(-1) == CivMap.GetNumResourcesOnLand(-1))
	assert(RsdMap.GetNumResources(-1) == CivMap.GetNumResources(-1))

	print("Testing GetNumImprovements");
	for rsdArea in RsdMap.Areas() do
		local civArea = CivMap.GetArea(rsdArea:GetID());
		for row in GameInfo.Improvements() do
			assert(rsdArea:GetNumImprovements(row.ID) == civArea:GetNumImprovements(row.ID))
		end
	end

	print("Testing FindBiggestArea");
	--print(RsdMap.FindBiggestArea():GetID(), RsdMap.FindBiggestArea():GetNumTiles(), CivMap.FindBiggestArea():GetID(), CivMap.FindBiggestArea():GetNumTiles(), tostring(#(RsdMap.inlinePlots)))
	assert(RsdMap.FindBiggestArea(true):GetID() == CivMap.FindBiggestArea(true):GetID())
	assert(RsdMap.FindBiggestArea(false):GetID() == CivMap.FindBiggestArea(false):GetID())

	print("Testing RecalculateAreas");
	CivMap.RecalculateAreas()
	RsdMap.RecalculateAreas()
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);
		--print(civPlot:GetX(), civPlot:GetY(), rsdPlot:Area():GetNumTiles(), civPlot:Area():GetNumTiles())
		assert(rsdPlot:Area():GetNumTiles() == civPlot:Area():GetNumTiles());
	end

	print("Restoring");
	DisableRsdMap()

	print("Ran all tests!");
	isRunningTests = false
end

-------------------------------------------------------------------------------------------------
local function GenerateRsdMapCore(script)
	print("Including: "..tostring(script));
	local result = include(script)

	if (not result) or (#result == 0) then 
		print(" Failure: "..tostring(script))
		return false 
	end

	print(" Success: "..result[1]);
	print("RESEED: Invoking GenerateMap");
	GenerateMap()
	print("RESEED: GenerateMap done");
	print("RESEED: Validating coast");
	RsdMap.ValidateCoast()
	return true
end

-------------------------------------------------------------------------------------------------
function RevealStartingPlot(plot)
	for dx = -3, 3, 1 do
		for dy = -3, 3, 1 do
			local neighbor = RsdMap.PlotXYWithRangeCheck(plot:GetX(), plot:GetY(), dx, dy, 3)
			if neighbor then 
				neighbor:SetRevealed(true) 
			end
		end
	end
end

-------------------------------------------------------------------------------------------------
function GenerateRsdMap()
	while true do
		EnableRsdMap()

		-- Invoke script
		local script = PreGame.GetMapScript();
		local scriptName = Path.GetFileName(script)
		local generated = GenerateRsdMapCore(script) or GenerateRsdMapCore(scriptName);
		if not generated then
			print("RESEED: Map generation error");
			DisableRsdMap()
			error("LOAD_MAP_ERROR")
		end

		print("RESEED: Map generation success");
		print("RESEED: Restoring map systems");
		DisableRsdMap()

		-- Ensure all civs have a starting position
		local allHaveStartingPosition = true
		for i = 0, GameDefines.MAX_CIV_PLAYERS do
			local player = Players[i];
			if player and player:IsEverAlive() and player:IsAlive() and not player:IsBarbarian() then
				if not RsdMap.GetStartingPlot(player) then
					allHaveStartingPosition = false
					break
				end
			end
		end

		if allHaveStartingPosition then break end
	end

	local plot = RsdMap.GetStartingPlot(Players[Game.GetActivePlayer()])
	RevealStartingPlot(plot)
	return RsdMap
end

-------------------------------------------------------------------------------------------------
function CopyCurrentMap()
	local playerPlot = nil

	-- Store goodies ID
	local goodies = {}
	for row in GameInfo.Improvements() do
		goodies[row.ID] = row.Goody
	end

	-- Collect players and start plots
	local starts = {}
	local players = {}
	for playerID = 0, GameDefines.MAX_CIV_PLAYERS do
		local player = Players[playerID];
		if player and player:IsEverAlive() and player:GetStartingPlot() then
			table.insert(starts, player:GetStartingPlot())
			players[player] = true
			if Game.GetActivePlayer() == playerID then
				playerPlot = player:GetStartingPlot()
			end
		end
	end

	-- Reset map
	EnableRsdMap()
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);

		rsdPlot:SetPlotType(civPlot:GetPlotType())
		rsdPlot:SetTerrainType(civPlot:GetTerrainType())
		rsdPlot:SetFeatureType(civPlot:GetFeatureType())
		rsdPlot:SetResourceType(civPlot:GetResourceType(), civPlot:GetNumResource())
		rsdPlot:SetContinentArtType(civPlot:GetContinentArtType())
		rsdPlot:SetWOfRiver(civPlot:IsWOfRiver(), civPlot:GetRiverEFlowDirection())
		rsdPlot:SetNWOfRiver(civPlot:IsNWOfRiver(), civPlot:GetRiverSEFlowDirection())
		rsdPlot:SetNEOfRiver(civPlot:IsNEOfRiver(), civPlot:GetRiverSWFlowDirection())
		rsdPlot:SetArea(civPlot:GetArea())

		if goodies[civPlot:GetImprovementType()] then
			rsdPlot:SetImprovementType(civPlot:GetImprovementType())
		end
	end

	-- Shuffle start plots
	local len = #starts;
	for i = 1, len, 1 do
		local k = Map.Rand(len - 1, "Shuffling Values") + 1;
		starts[i], starts[k] = starts[k], starts[i];
	end
	-- Second pass to ensure the player's position is indeed reset (the user could think of a bug)
	if starts[1] == playerPlot then
		local k = Map.Rand(len - 2, "Shuffling Values") + 2;
		starts[1], starts[k] = starts[k], starts[1];
	end

	-- Assign start plots
	local startIndex = 2
	for player in pairs(players) do
		local plot
		if player:GetID() == Game.GetActivePlayer() then 
			plot = starts[1]
		else
			plot = starts[startIndex]
			startIndex = startIndex + 1
		end
		player:SetStartingPlot(plot);
	end

	RsdMap.CalculateAreas();
	DisableRsdMap()
	return RsdMap
end

--[[local abc = Map.GetPlotByIndex(i)
abc:SetTerrainType(1)
abc:SetTerrainType(nil)
print("abc="..sanitize(abc:GetTerrainType())) -- abc is now 0]]

--[[local abc = Map.GetPlotByIndex(i)
abc:SetPlotType(1)
abc:SetPlotType(nil)
print("abc="..sanitize(abc:GetPlotType())) -- abc is now 0]]

--RunTests()
