-- Released under GPL v3
--------------------------------------------------------------
-- Increase this setting to increase speed. User lower values if you observe crashes. Relevant range is 1 to 50000 (both included)
local BatchSize = 100
local Settings = 
{
	-- Safest
	{	InitUnitWeight = 50, MoveUnitWeight = 50, DestroyUnitWeight = 50, DestroyCityWeight = 100, PlayerWeight = 10,
		TerrainWeight = 1, FogWeight = 1, TechnoWeight = 5, PolicyWeight = 5, DeletionDelay = true	},

	-- Balanced
	{	InitUnitWeight = 10, MoveUnitWeight = 10, DestroyUnitWeight = 10, DestroyCityWeight = 100, PlayerWeight = 5,
		TerrainWeight = 0.2, FogWeight = 0.2, TechnoWeight = 1, PolicyWeight = 1, DeletionDelay = false	},

	-- Fast
	{	InitUnitWeight = 1, MoveUnitWeight = 1, DestroyUnitWeight = 1, DestroyCityWeight = 100, PlayerWeight = 1,
		TerrainWeight = 0.2, FogWeight = 0.2, TechnoWeight = 0.2, PolicyWeight = 0.2, DeletionDelay = false	},

	-- Fastest
	{	InitUnitWeight = 1, MoveUnitWeight = 1, DestroyUnitWeight = 1, DestroyCityWeight = 50, PlayerWeight = 1,
		TerrainWeight = 0, FogWeight = 0, TechnoWeight = 0, PolicyWeight = 0, DeletionDelay = false	},
}

local beaconID = GameInfo.Units.RSD_BEACON.ID
local unitNames = {}
local queue = {}

for row in GameInfo.Units() do
	unitNames[row.ID] = row.Type
end

local options = {};
LuaEvents.RSD_ShareOptions.Add(function(_options) options = _options end)

-------------------------------------------------------------------------------------------------
local function PrintUnits(players)
	print("Checking units")
	for player in pairs(players) do
		print(" "..player:GetName(), player:GetNumUnits())
		--[[
		local foundAny = false
	    for unit in player:Units()
			if unit then
				foundAny = true
				print("  found", unitNames[unit:GetUnitType()], unit:GetX(), unit:GetY())
			end
		end
		if not foundAny then
			print("  no units")
		end]]
	end
end

--===============================================================================================
-- STARTING UNITS SPAWNS
--===============================================================================================
local function GetDefaultUnitForClass(classType, civType)
	-- Specific hack for shoshone since Firaxis did not use the standard mechanics but relied on a built-in hack.
	if civType == "CIVILIZATION_SHOSHONE" and classType == "UNITCLASS_WARRIOR" then 
		return GameInfo.Units.UNIT_SHOSHONE_PATHFINDER.ID
	end

	if civType then
		for row in GameInfo.Civilization_UnitClassOverrides{ CivilizationType=civType, UnitClassType=classType} do
			if row.UnitType then
				local unit = GameInfo.Units[row.UnitType];
				if unit then
					return unit.ID;
				end
			end
		end
	end

	local classRow = GameInfo.UnitClasses[classType];
	if classRow then
		local unitRow = GameInfo.Units[classRow.DefaultUnit];
		return unitRow.ID;
	end

	print("invalid class: "..tostring(classType));
	return 0;
end

-------------------------------------------------------------------------------------------------
local function GetStartingUnits(player)
	local units = {};
	local civType = GameInfo.Civilizations[player:GetCivilizationType()].Type;
	local handicap = GameInfo.HandicapInfos[Game.GetHandicapType()];
	local era = GameInfo.Eras[Game.GetStartEra()]

	local workerCount = 0;
	local defenseCount = 0;
	local explorerCount = 0;
	local freeUnitCountMultiplier = 1;
	if (not player:IsHuman()) and (not player:IsMinorCiv()) then
	end

	if player:IsHuman() then
		workerCount = era.StartingWorkerUnits + handicap.StartingWorkerUnits;
		defenseCount = era.StartingDefenseUnits + handicap.StartingDefenseUnits;
		explorerCount = era.StartingExploreUnits + handicap.StartingExploreUnits;
		freeUnitCountMultiplier = era.StartingUnitMultiplier;
	elseif not player:IsMinorCiv() then
		workerCount = era.StartingWorkerUnits + handicap.AIStartingWorkerUnits;
		defenseCount = era.StartingDefenseUnits + handicap.AIStartingDefenseUnits;
		explorerCount = era.StartingExploreUnits + handicap.AIStartingExploreUnits;
		freeUnitCountMultiplier = era.StartingUnitMultiplier + handicap.AIStartingUnitMultiplier;
	end

	-- Add free units (settlers)
	for row in GameInfo.Civilization_FreeUnits("CivilizationType='"..civType.."'") do
		local freeUnit = GetDefaultUnitForClass(row.UnitClassType, civType);
		local freeUnitCount = row.Count * freeUnitCountMultiplier;

		for i = 0, freeUnitCount - 1, 1 do
			table.insert(units, freeUnit);
		end
	end

	-- Add defense units
	local defenseUnit = GetDefaultUnitForClass("UNITCLASS_WARRIOR", civType);
	for i = 0, defenseCount - 1, 1 do
		table.insert(units, defenseUnit);
	end

	-- Add explorer units
	local explorerUnit = GetDefaultUnitForClass("UNITCLASS_SCOUT", civType);
	for i = 0, explorerCount - 1, 1 do
		table.insert(units, explorerUnit);
	end

	-- Add worker units
	local workerUnit = GetDefaultUnitForClass("UNITCLASS_WORKER", civType);
	for i = 0, workerCount - 1, 1 do
		table.insert(units, workerUnit);
	end

	return units;
end

-------------------------------------------------------------------------------------------------
local function GetSpawnLocations(startPlot, units)
	-- Store land plots around starting position
	local plots = { startPlot };
	for i = 0, 5, 1 do
		local plot = RsdMap.PlotDirection(startPlot:GetX(), startPlot:GetY(), i);
		if plot then
			local type = plot:GetTerrainType();
			local plotType = plot:GetPlotType();
			if (plotType ~= PlotTypes.PLOT_MOUNTAIN) and (not plot:IsWater()) and (not plot:IsImpassable()) then
				table.insert(plots, plot);
			end
		end
	end

	-- Spawn units
	local results = {}
	local plotIndex = 1;
	for j, v in ipairs(units) do
		if j > 0 then
			local plot = plots[plotIndex];
			plotIndex = plotIndex + 1;
			if plotIndex > #plots then plotIndex = 2 end
			if #plots == 0 then plot = startPlot end	-- Poor player, a single hex of land... :(

			table.insert(results, { x = plot:GetX(), y = plot:GetY(), unit = v })
		end
	end
	return results
end

--===============================================================================================
-- INTERFERENCE-FREE UNITS MOVES
--===============================================================================================
local function Increment(t, key)
	t[key] = t[key] and t[key] + 1 or 1
end

local function Decrement(t, key)
	t[key] = t[key] - 1
end

-------------------------------------------------------------------------------------------------
local function CountInterferences(plot, blocking, isWithin)
	local count = 0
	for otherPlot, value in pairs(blocking) do
		if value and value > 0 then
			if isWithin and (plot == otherPlot) then
				if value > 1 then count = count + 1 end
			elseif Map.PlotDistance(plot:GetX(), plot:GetY(), otherPlot:GetX(), otherPlot:GetY()) < 5 then
				count = count + 1
			end
		end
	end
	return count
end

-------------------------------------------------------------------------------------------------
local function CanMoveTo(plot, blocking, isWithin)
	return CountInterferences(plot, blocking, isWithin) == 0
end

-------------------------------------------------------------------------------------------------
local function StoreMove(player, plot, current, free, all, moves, isFinal)
	local old = current[player]

	table.insert(moves, { player = player, x = plot:GetX(), y = plot:GetY(), isFinal = isFinal })
	player.isOnStart = false
	current[player] = plot
	free[plot] = nil			-- No more free

	-- Is the old location a possible intermediate site?
	if CanMoveTo(old, all, true) then
		--print("make it free")
		free[old] = true		-- Make it free
	else
		--print("delete old")
		Decrement(all, old)		-- Just remove it from the list
	end
end

-------------------------------------------------------------------------------------------------
local function TrySwap(player, current, targets, free, all, moves)
	local target = targets[player]
	assert(target)

	-- Try to move to the target
	if CanMoveTo(target, all, true) then
		StoreMove(player, target, current, free, all, moves, true)
		current[player] = nil
		targets[player] = nil
		return 2
	end

	-- Player already moved to a free site?
	if not player.isOnStart then return false end

	-- Move to some free site if still on original location
	local firstFree
	for k, v in pairs(free) do
		if v then 
			firstFree = k
			break
		end
	end
	if not firstFree then return false end
	StoreMove(player, firstFree, current, free, all, moves)
	return 1
end

-------------------------------------------------------------------------------------------------
local function GetUnitSwaps(players)
	print("GetUnitSwaps")
	local inlineTargets = {}
	local current = {}
	local targets = {}
	local moves = {}
	local free = {}
	local all = {}

	-- Build tables
	print(" compute all")
	for player in pairs(players) do
		local oldPlot = player:GetStartingPlot()
		local newRsdPlot = RsdMap.GetStartingPlot(player)
		local newPlot = Map.GetPlot(newRsdPlot:GetX(), newRsdPlot:GetY())

		current[player] = oldPlot
		targets[player] = newPlot
		Increment(inlineTargets, newPlot)
		Increment(all, oldPlot)		-- a current plot could be the same as one of the targets, hence the increment/decrement
		Increment(all, newPlot)
		player.isOnStart = true

		print("  "..player:GetName(), oldPlot:GetX(), oldPlot:GetY())
		print("  "..player:GetName(), newPlot:GetX(), newPlot:GetY())
	end

	-- Sort players according to which one is blocking the most
	print(" compute interferences")
	local sortedPlayers = {}
	for player in pairs(players) do
		assert(current[player])
		player.interferences = CountInterferences(current[player], inlineTargets, false)
		table.insert(sortedPlayers, player)
	end
	table.sort(sortedPlayers, function(a, b) return a.interferences > b.interferences end)

	for _, v in ipairs(sortedPlayers) do
		print("  "..v:GetName(), v.interferences)
	end

	-- Build free spots table
	print(" compute free plots")
	for i = 0, Map.GetNumPlots() - 1, 1 do
		local plot = Map.GetPlotByIndex(i)
		if CanMoveTo(plot, all) then
			print("  "..plot:GetX(), plot:GetY())
			Increment(all, plot)
			free[plot] = true
		end
	end

	-- Move players
	print(" perform moves")
	while true do
		local i = 1
		local movedAny = false
		print("  new swap pass ("..#sortedPlayers.." left)")
		while i <= #sortedPlayers do
			local player = sortedPlayers[i]
			local result = TrySwap(player, current, targets, free, all, moves)

			if not result then
				print("   skipped ", player:GetName())
				i = i + 1
			elseif result == 1 then
				print("   moved ", player:GetName())
				movedAny = true
				i = i + 1
			else
				print("   done ", player:GetName())
				table.remove(sortedPlayers, i)
				movedAny = true
			end
		end

		if #sortedPlayers == 0 then return moves end
		if not movedAny then error("MOVE_ERROR") end
	end
end

--===============================================================================================
-- MAP GENERATION PROCESS CORE
--===============================================================================================
local startTime;
local completeKills;
local dominationVictory;
local currentBatchSize;

-------------------------------------------------------------------------------------------------
local function Yield(weight)
	currentBatchSize = currentBatchSize + (weight or 1)
	if currentBatchSize > BatchSize then coroutine.yield() end
end

-------------------------------------------------------------------------------------------------
local function Status(str)
	print(str)
	LuaEvents.RSD_NewMapStatus(L(str))
	coroutine.yield()
end

-------------------------------------------------------------------------------------------------
local function DumpStopwatch()
	local t2 = os.clock()
	local dt = math.ceil((t2 - startTime) * 1000);
	print("Done in "..dt.."ms")
end

-------------------------------------------------------------------------------------------------
local function NewMap(map, useCurrentMap)
	local teams = {}
	local players = {}
	local techsByTeam = {}
	local startingUnits = {}
	local startingSpawns = {}
	local processedPlots = {}
	local playersButBarbarians = {}
	local startingEra = PreGame.GetEra()
	local settings = Settings[options.speed]

	print("====================BEGIN==========================")
	print("Player: "..Players[Game.GetActivePlayer()]:GetName());
	startTime = os.clock()

	local notificationHandler = function(id) UI.RemoveNotification(id) end
	Events.NotificationAdded.Add(notificationHandler);
	UI.RebroadcastNotifications();

	Status("TXT_KEY_RSD_STEP_VIRTUAL_MAP")
	if useCurrentMap then
		RsdMap = CopyCurrentMap()
	elseif map then
		RsdMap = map
	else
		RsdMap = GenerateRsdMap()
	end

	Status("TXT_KEY_RSD_STEP_HIDE_MAP")
	FOW_SetAll(0);
	FOW_SetAll(1);

	Status("TXT_KEY_RSD_STEP_GATHER_DATA")
	for i = 0, GameDefines.MAX_CIV_PLAYERS do
		local player = Players[i];
		if player and player:IsEverAlive() and (player:IsAlive() or player:IsBarbarian()) then
			local teamID = player:GetTeam();
			players[player] = true
			teams[teamID] = true
			techsByTeam[teamID] = {}

			if not player:IsBarbarian() then
				local rsdPlot = RsdMap.GetStartingPlot(player)
				local civPlot = Map.GetPlot(rsdPlot:GetX(), rsdPlot:GetY())
				startingUnits[player] = GetStartingUnits(player)
				startingSpawns[player] = GetSpawnLocations(rsdPlot, startingUnits[player])
				playersButBarbarians[player] = true
			end
		end
	end
	local swaps = GetUnitSwaps(playersButBarbarians)
	PrintUnits(players)

	Status("TXT_KEY_RSD_STEP_FLUSH_PLAYERS")
	Game.SetGameTurn(Game.GetStartTurn());
	for player in pairs(players) do
		player:SetGold(0);
		player:SetJONSCulture(0);
		player:SetAnarchyNumTurns(0);
		player:SetCombatExperience(0);
		player:SetConscriptCount(0);
		player:SetNumFreePolicies(0);
		player:SetNumFreeTechs(0);
		player:SetNumGoldenAges(0);
		player:SetNumPlotsBought(0);
		player:SetResearchingTech(0);
		if player.SetFaith then
			player:SetFaith(0);
		end
		if player.SetNumFreeTenets then
			player:SetNumFreeTenets(0);
		end
		Yield(settings.PlayerWeight * 11)
	end

	Status("TXT_KEY_RSD_STEP_FLUSH_POLICIES")
	for player in pairs(players) do
		for row in GameInfo.Policies() do
			player:SetHasPolicy(row.ID, false);
			Yield(settings.PolicyWeight)
		end
		for row in GameInfo.PolicyBranchTypes() do
			player:SetPolicyBranchUnlocked(row.ID, false);
			Yield(settings.PolicyWeight * 5)
		end
	end



	Status("TXT_KEY_RSD_STEP_FLUSH_TECHS")
	-- Collect free eras
	local freeEras = {}
	for row in GameInfo.Eras() do
		freeEras[row.Type] = (row.ID < startingEra)
	end

	-- Collect free techs
	for player in pairs(players) do
		local teamID = player:GetTeam()
		local civType = GameInfo.Civilizations[player:GetCivilizationType()].Type;
		for row in GameInfo.Civilization_FreeTechs{CivilizationType=civType} do
			if row.TechType then
				techsByTeam[teamID][row.TechType] = true;
			end
		end
	end

	-- Modify techs status
	for teamID, freeTechs in pairs(techsByTeam) do
		local team = Teams[teamID];
		for row in GameInfo.Technologies() do
			if row.ID and row.Type and row.Era then
				local isGranted = freeTechs[row.Type] or freeEras[row.Era]
				local bValue = isGranted and true or false;
				team:SetHasTech(row.ID, bValue, player, true, true);
				Yield(settings.TechnoWeight)
			end
		end
		team:SetCurrentEra(startingEra)
	end


	Status("TXT_KEY_RSD_STEP_SPAWN_BEACONS")
	for player in pairs(playersButBarbarians) do
		local plot = player:GetStartingPlot()
		player:InitUnit(beaconID, plot:GetX(), plot:GetY())
		print(" spawned ", player:GetName(), beaconID, plot:GetX(), plot:GetY())
		Yield(settings.InitUnitWeight)
	end

	Status("TXT_KEY_RSD_STEP_FLUSH_CITIES")
	PrintUnits(playersButBarbarians)
	for player in pairs(players) do
		for i = 0, player:GetNumCities() - 1, 1 do
			local city = player:GetCityByID(i);
			if city then
				city:Kill(settings.DeletionDelay)
				Yield(settings.DestroyCityWeight)
			end
		end
	end

	Status("TXT_KEY_RSD_STEP_FLUSH_UNITS")
	for player in pairs(players) do
		for unit in player:Units() do
			if unit and unit:GetUnitType() ~= beaconID and not unit:IsDead() then
				unit:Kill(settings.DeletionDelay)
				Yield(settings.DestroyUnitWeight)
			end
		end
	end

	-- Steps below ensure that units won't encounter natural wonders, goodies or other civs' territory. Also clean up a little.
	Status("TXT_KEY_RSD_STEP_CLEAN_MAP_1")
	PrintUnits(playersButBarbarians)
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		civPlot:SetFeatureType(-1)
		civPlot:SetImprovementType(-1)
		civPlot:SetImprovementPillaged(false)
		civPlot:SetOwner(-1)
		civPlot:SetRouteType(-1)
		Yield(settings.TerrainWeight * 5)
	end

	-- Steps below ensure that units won't be moved or spawned on ocean/mountain tiles
	Status("TXT_KEY_RSD_STEP_CLEAN_MAP_2")
	local plotsToCleanEarly = {}
	for _, v in ipairs(swaps) do
		table.insert(plotsToCleanEarly, v)
	end
	for player in pairs(playersButBarbarians) do
		for _, v in ipairs(startingSpawns[player]) do
			table.insert(plotsToCleanEarly, v)
		end
	end

	local invalidPlotTypes = {};
	invalidPlotTypes[PlotTypes.PLOT_OCEAN] = true
	invalidPlotTypes[PlotTypes.PLOT_MOUNTAIN] = true
	for _, v in ipairs(plotsToCleanEarly) do
		local civPlot = CivMap.GetPlot(v.x, v.y);
		if invalidPlotTypes[civPlot:GetPlotType()] or civPlot:IsImpassable() then
			civPlot:SetPlotType(PlotTypes.PLOT_LAND)
			civPlot:SetTerrainType(TerrainTypes.TERRAIN_PLAINS)
			Yield(settings.TerrainWeight * 2)
		end
	end

	Status("TXT_KEY_RSD_STEP_MOVE_UNITS")
	PrintUnits(playersButBarbarians)
	for _, v in ipairs(swaps) do
		for unit in v.player:Units() do
			if unit and not unit:IsDead() then
				unit:SetXY(v.x, v.y)
				print(" "..v.player:GetName(), "moved", unitNames[unit:GetUnitType()], unit:GetX(), unit:GetY(), v.x, v.y)
				Yield(settings.MoveUnitWeight)
			end
		end
	end

	Status("TXT_KEY_RSD_STEP_START_PLOTS")
	for player in pairs(playersButBarbarians) do
		local rsdPlot = RsdMap.GetStartingPlot(player)
		local civPlot = Map.GetPlot(rsdPlot:GetX(), rsdPlot:GetY())
		player:SetStartingPlot(civPlot)
	end

	Status("TXT_KEY_RSD_STEP_SPAWN_UNITS")
	for player in pairs(playersButBarbarians) do
		for _, v in ipairs(startingSpawns[player]) do
			player:InitUnit(v.unit, v.x, v.y)
			print(" spawned ", player:GetName(), unitNames[v.unit], tostring(v.x), tostring(v.y))
			Yield(settings.InitUnitWeight)
		end
	end

	Status("TXT_KEY_RSD_STEP_FLUSH_BEACONS")
	for player in pairs(playersButBarbarians) do
		for unit in player:Units() do
			if unit and unit:GetUnitType() == beaconID and not unit:IsDead() then
				unit:Kill(settings.DeletionDelay)
				Yield(settings.DestroyUnitWeight)
			end
		end
	end

	Status("TXT_KEY_RSD_STEP_UPDATE_FOG")
	PrintUnits(playersButBarbarians)
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);

		for teamID in pairs(teams) do
			civPlot:SetRevealed(teamID, false)
			Yield(settings.FogWeight)
		end
		--civPlot.UpdateFog()
	end

	Status("TXT_KEY_RSD_STEP_UPDATE_MAP")
	for i = 0, Map.GetNumPlots()-1, 1 do
		local civPlot = CivMap.GetPlotByIndex(i);
		local rsdPlot = RsdMap.GetPlotByIndex(i);
		civPlot:SetPlotType(rsdPlot:GetPlotType())
		civPlot:SetTerrainType(rsdPlot:GetTerrainType())
		civPlot:SetFeatureType(rsdPlot:GetFeatureType())
		civPlot:SetImprovementType(rsdPlot:GetImprovementType())
		civPlot:SetContinentArtType(rsdPlot:GetContinentArtType())
		civPlot:SetResourceType(rsdPlot:GetResourceType(), rsdPlot:GetNumResource())
		civPlot:SetNWOfRiver(rsdPlot:IsNWOfRiver(), rsdPlot:GetRiverSEFlowDirection())
		civPlot:SetNEOfRiver(rsdPlot:IsNEOfRiver(), rsdPlot:GetRiverSWFlowDirection())
		civPlot:SetWOfRiver(rsdPlot:IsWOfRiver(), rsdPlot:GetRiverEFlowDirection())
		Yield(settings.TerrainWeight * 9)
	end

	Status("TXT_KEY_RSD_STEP_UPDATE_AREAS")
	Map.RecalculateAreas()

	Status("TXT_KEY_RSD_STEP_FINALIZE_MAP")
	if RsdMap.useDefaultContinentStamper then
		Map.DefaultContinentStamper()
	end
	Map.ChangeAIMapHint(RsdMap.mapHint)
	DumpStopwatch()

	-- Support for "ResourceGenerator" (mod base for new resources)
	local resGenData = Modding.OpenSaveData()
	resGenData.SetValue("Resources_Generation_Completed", 0);

	Status("TXT_KEY_RSD_STEP_SAVE")
	Events.NotificationAdded.Remove(notificationHandler);
	--DeleteFiles("__reseed")
	UI.SaveGame("__reseed", true)

	-- Clean up may reduce crashes (doubtful but one user reported it)
	RsdMap = nil;
	local settings = nil
	local startingEra = nil
	local playersButBarbarians = nil
	local processedPlots = nil
	local startingSpawns = nil
	local startingUnits = nil
	local techsByTeam = nil
	local players = nil
	local teams = nil

	Status("TXT_KEY_RSD_STEP_RELOAD")
	coroutine.yield()	-- Waiting may reduce crashes for users plagued with them.
	coroutine.yield()
	coroutine.yield()
	LoadFile("__reseed")
end

--===============================================================================================
-- MAP GENERATION PROCESS SETUP
--===============================================================================================
local function ResumeCoroutine(newMapCoroutine, ...)
	currentBatchSize = 0;
	print("invoke")
	if newMapCoroutine and coroutine.status(newMapCoroutine) ~= "dead" then
		local result, err, returned = coroutine.resume(newMapCoroutine, ...)
		if not result then
			print("Coroutine error");
			print(tostring(result), tostring(err));
			
			if string.find(err, "MOVE_ERROR", 1, true) then
				err = "MOVE_ERROR"
			elseif string.find(err, "LOAD_MAP_ERROR", 1, true) then
				err = "LOAD_MAP_ERROR"
			else
				err = FormatError(err, 0, newMapCoroutine);
			end

			LuaEvents.RSD_RoutineError(err)
			DisableRsdMap()
			return "error"
		elseif coroutine.status(newMapCoroutine) == "dead" then
			return "done"
		else
			return "yield"
		end
	end
	return "halted"
end

-------------------------------------------------------------------------------------------------
function CreateNewMapCoroutine(map, useCurrentMap)
	local co = coroutine.create(NewMap)
	return function() return ResumeCoroutine(co, map, useCurrentMap) end
end

-------------------------------------------------------------------------------------------------
function CreateCoroutine(callback)
	local co = coroutine.create(callback)
	print(coroutine.status(co))
	return function() return ResumeCoroutine(co) end
end
