-- Released under GPL v3
--------------------------------------------------------------
include("IconSupport");
include("InstanceManager");
include("RSD_Utils");
local riverManager = InstanceManager:new("RsdRiverInstance", "Anchor", Controls.RiverContainer);
local hexManager = InstanceManager:new("RsdHexInstance", "Anchor", Controls.HexContainer);
local fogManager = InstanceManager:new("RsdFogInstance", "Anchor", Controls.FogContainer);
local riverInstances = {}
local hexInstances = {}
local fogInstances = {}

local options = {};
LuaEvents.RSD_ShareOptions.Add(function(_options) options = _options end)

local terrains = {}
terrains[TerrainTypes.TERRAIN_COAST] = "RsdCoast.dds";
terrains[TerrainTypes.TERRAIN_OCEAN] = "RsdOcean.dds";
terrains[TerrainTypes.TERRAIN_GRASS] = "RsdGrass.dds";
terrains[TerrainTypes.TERRAIN_PLAINS] = "RsdPlains.dds";
terrains[TerrainTypes.TERRAIN_DESERT] = "RsdDesert.dds";
terrains[TerrainTypes.TERRAIN_TUNDRA] = "RsdTundra.dds";
terrains[TerrainTypes.TERRAIN_SNOW] = "RsdSnow.dds";

local types = {}
types[PlotTypes.PLOT_LAND] = {}
types[PlotTypes.PLOT_HILLS] = {}
types[PlotTypes.PLOT_MOUNTAIN] = {}
types[PlotTypes.PLOT_OCEAN] = {}

types[PlotTypes.PLOT_HILLS][-1] = "sv_hills.dds"
types[PlotTypes.PLOT_HILLS][TerrainTypes.TERRAIN_SNOW] = "sv_hills_on_snow.dds"
types[PlotTypes.PLOT_HILLS][TerrainTypes.TERRAIN_TUNDRA] = "sv_hills_on_tundra.dds"

types[PlotTypes.PLOT_MOUNTAIN][-1] = "sv_mountains.dds"
types[PlotTypes.PLOT_MOUNTAIN][TerrainTypes.TERRAIN_SNOW] = "sv_mountains_on_snow.dds"

local features = {};
local isWonder = {}
for row in GameInfo.Features() do
	features[row.ID] = {}
	isWonder[row.ID] = row.NaturalWonder;
	if row.ArtDefineTag and row.ArtDefineTag ~= "" then
		for subRow in GameInfo.ArtDefine_StrategicView{StrategicViewType=row.ArtDefineTag} do
			features[row.ID][-1] = subRow.Asset;
		end
	end
end
--features[FeatureTypes.FEATURE_FOREST][TerrainTypes.TERRAIN_PLAINS] = "sv_forest_on_plains.dds";
features[-1] = {};

local resources = {};
for row in GameInfo.Resources() do
	if row.ArtDefineTag and row.ArtDefineTag ~= "" then
		for subRow in GameInfo.ArtDefine_StrategicView{StrategicViewType=row.ArtDefineTag} do
			resources[row.ID] = subRow.Asset;
		end
	end
end

local improvements = {};
for row in GameInfo.Improvements() do
	if row.ArtDefineTag and row.ArtDefineTag ~= "" then
		for subRow in GameInfo.ArtDefine_StrategicView{StrategicViewType=row.ArtDefineTag} do
			improvements[row.ID] = subRow.Asset;
		end
	end
end

local Map = nil;
local revealMap = false
local startingPlotGetter = nil;

-------------------------------------------------------------------------------------------------
-- VERTICES	  		 SIDES
--     0			   .
--	 /	 \			 0   5
--  1     5	       .       .
--	|	  |		   1       4
--  2     4		   .       .
--	 \	 /			 2   3
--     3               .
-------------------------------------------------------------------------------------------------
local function HasRiverOnSide(plot, index)
	index = (index + 6 ) % 6
	if index == 0 then -- NW
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHWEST)
		return other and other:IsNWOfRiver()
	elseif index == 1 then -- W
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_WEST)
		return other and other:IsWOfRiver()
	elseif index == 2 then -- SW
		return plot:IsNEOfRiver()
	elseif index == 3 then -- SE
		return plot:IsNWOfRiver()
	elseif index == 4 then -- E
		return plot:IsWOfRiver()
	else -- NE
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHEAST)
		return other and other:IsNEOfRiver()
	end
end

-------------------------------------------------------------------------------------------------
local function HasVertexOutgoingRiver(plot, index)
	index = (index + 6 ) % 6
	if index == 0 then -- N
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHWEST)
		return other and other:IsWOfRiver();
	elseif index == 1 then -- NW
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHWEST)
		return other and other:IsNEOfRiver();
	elseif index == 2 then -- SW
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_WEST)
		return other and other:IsNWOfRiver();
	elseif index == 3 then -- S
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST)
		return other and other:IsWOfRiver();
	elseif index == 4 then -- SE
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_EAST)
		return other and other:IsNEOfRiver();
	else -- NE
		local other = Map.PlotDirection(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHEAST)
		return other and other:IsNWOfRiver()
	end
end

-------------------------------------------------------------------------------------------------
local function IsNeighborWater(plot, direction)
	local other = Map.PlotDirection(plot:GetX(), plot:GetY(), direction)
	return other and other:IsWater();
end

-------------------------------------------------------------------------------------------------
-- Returns 0 if no river should be displayed (may have a river displayed by another plot)
-- Returns "x3" if a three-edges connection should be displayed 
-- Returns "x2" if a two-edges connection should be displayed \__/
-- Returns "cws"  if a river source is present in CW direction
-- Returns "ccws" if a river source is present in CCW direction
-- Returns "cwe"  if a river ending is present in CW direction
-- Returns "ccwe" if a river ending is present in CCW direction
-------------------------------------------------------------------------------------------------
local vertexCWDirection = {};
vertexCWDirection[0] = DirectionTypes.DIRECTION_NORTHWEST;
vertexCWDirection[1] = DirectionTypes.DIRECTION_WEST;
vertexCWDirection[2] = DirectionTypes.DIRECTION_SOUTHWEST;
vertexCWDirection[3] = DirectionTypes.DIRECTION_SOUTHEAST;
vertexCWDirection[4] = DirectionTypes.DIRECTION_EAST;
vertexCWDirection[5] = DirectionTypes.DIRECTION_NORTHEAST;

local function GetRiverStatusForVertex(plot, index)
	index = (index + 6 ) % 6
	local after = HasRiverOnSide(plot, index)
	local before = HasRiverOnSide(plot, index - 1)
	local outgoing = HasVertexOutgoingRiver(plot, index)

	-- No river on the edges
	if (not after) and (not before) then return 0 end

	-- Three-way connection
	if after and before and outgoing then 
		if index > 1 then return end
		return "x3" 
	end

	-- Two-way connection
	if after and before then return "x2" end

	-- River on one side only, continued in another plot: this other plot will display it, do nothing now
	if outgoing then return 0 end

	-- River stops here, is it a source or an ending?
	if before then
		if index > 2 then return 0 end
		return IsNeighborWater(plot, vertexCWDirection[index]) and "cwe" or "cws";
	else
		if index > 1 and index ~= 5 then return 0 end
		return IsNeighborWater(plot, vertexCWDirection[(index + 5) % 6]) and "ccwe" or "ccws";
	end
end

-- Set coordinates for RsdRivers.dds. Look at this file first to understand.
-------------------------------------------------------------------------------------------------

-- First store the position to align sources on vertices. Convention: during a CW rotation, the ending is the head and the source the tail
local riverSources = {}		
riverSources[0] = { x = 149,	y = 751 }
riverSources[1] = { x = 450,	y = 452 }
riverSources[2] = { x = 755,	y = 450 }
riverSources[3] = { x = 150,	y = 449 }
riverSources[4] = { x = 451,	y = 752 }
riverSources[5] = { x = 754,	y = 150 }

-- Compute endings from sources
local riverEndings = {}
for i = 0, 5, 1 do
	local angle = (2 * i - 1) * math.pi / 6;
	local c = math.cos(angle);
	local s = math.sin(angle);
	local r = 115;

	riverEndings[i] = { 
		x = riverSources[i].x + r * c, 
		y = riverSources[i].y - r * s 
	};
end

-- Assign all coordinates
local riverTextures = {};
local riverImagesLocations = {};
for i = 0, 5, 1 do 
	local angle = (2 * i + 3) * math.pi / 6;
	local c = math.cos(angle);
	local s = math.sin(angle);

	riverTextures[i] = {};
	riverImagesLocations[i] = { x = 208 * c,	y = -208 * s }

	-- Three-edges intersection
	if (i % 2) == 0 then
		riverTextures[i].x3 = { x = 151,		y = 147 }
	else
		riverTextures[i].x3 = { x = 453,		y = 152 }
	end

	-- Two edges intersection
	riverTextures[i].x2 = { x = 750 + 125 * c,	y = 750 - 125 * s };

	-- Sources and endings
	riverTextures[i].cws =  riverSources[i]
	riverTextures[i].ccwe = riverEndings[(i + 1) % 6]
	riverTextures[i].cwe =  riverEndings[(i + 3) % 6]
	riverTextures[i].ccws = riverSources[(i + 4) % 6]
end

-- Fix coordinates: so far we provided centers but we need their top left corners.
local processed = {}
for i = 0, 5, 1 do
	for _, v in pairs(riverTextures[i]) do
		if not processed[v] then
			processed[v] = true;
			v.x = v.x - 70		
			v.y = v.y - 70		
		end
	end
end


-------------------------------------------------------------------------------------------------
local resourcesVisibility = {}
local function UpdateResourceVisibility()
	local player = Players[Game.GetActivePlayer()];
	local team = Teams[player:GetTeam()];

	for row in GameInfo.Resources() do
		if options.showUnknownResources or (not row.TechReveal) then
			resourcesVisibility[row.ID] = true
		else
			local tech = GameInfoTypes[row.TechReveal];
			resourcesVisibility[row.ID] = team:GetTeamTechs():HasTech(tech);
		end
	end
end

-------------------------------------------------------------------------------------------------
local startingPlots = {};
local function UpdatePlot(plot, instance, riverInstance, fogInstance)
	local visible = revealMap or plot:IsRevealed(Game.GetActiveTeam())
	fogInstance.Fog:SetHide(visible)
	instance.Civ:SetHide(not visible)
	instance.Type:SetHide(not visible)
	instance.Terrain:SetHide(not visible)
	instance.Feature:SetHide(not visible)
	instance.Resource:SetHide(not visible)
	instance.Improvement:SetHide(not visible)

	if visible then
		instance.Terrain:SetTexture(terrains[plot:GetTerrainType()])

		-- Hills / moutains
		if not isWonder[plot:GetFeatureType()] then
			local typeTexture = types[plot:GetPlotType()][plot:GetTerrainType()] or types[plot:GetPlotType()][-1];
			instance.Type:SetHide(not typeTexture);
			instance.Type:SetTexture(typeTexture);
		else
			instance.Type:SetHide(true);
		end

		-- Features and natural wonders
		local featureTexture = features[plot:GetFeatureType()][plot:GetTerrainType()] or features[plot:GetFeatureType()][-1];
		instance.Feature:SetHide(not featureTexture);
		instance.Feature:SetTexture(featureTexture);

		-- Resources
		if resourcesVisibility[plot:GetResourceType()] then
			local resourceTexture = resources[plot:GetResourceType()];
			instance.Resource:SetHide(not resourceTexture);
			instance.Resource:SetTexture(resourceTexture);
		else
			instance.Resource:SetHide(true);
		end

		-- Improvement
		if options.showImprovements then
			local improvementTexture = improvements[plot:GetImprovementType()];
			instance.Improvement:SetHide(not improvementTexture);
			instance.Improvement:SetTexture(improvementTexture);
		else
			instance.Improvement:SetHide(true)
		end

		-- Starting location
		local player = startingPlots[plot];
		if player then
			local civ = GameInfo.Civilizations[player:GetCivilizationType()]
			local textureOffset, texture;
			if player:IsMinorCiv() then
				textureOffset, texture = IconLookup(civ.PortraitIndex, 256, civ.AlphaIconAtlas);	
			else
				textureOffset, texture = IconLookup(civ.PortraitIndex, 256, civ.IconAtlas);	
			end

			instance.Civ:SetTextureOffset(textureOffset)
			instance.Civ:SetTexture(texture)
			instance.Civ:SetHide(false)
		else
			instance.Civ:SetHide(true)
		end
	end

	-- Rivers
	for i = 0, 5, 1 do
		local ctl = riverInstance["Vertex"..i];
		local status = GetRiverStatusForVertex(plot, i);
		local texCoords = riverTextures[i][status];
		if texCoords then
			--print(plot:GetX(), plot:GetY(), i, status);
			ctl:SetTextureOffsetVal(riverTextures[i][status].x, riverTextures[i][status].y)
		end
		ctl:SetHide(status == 0);
	end
end

-------------------------------------------------------------------------------------------------
local function Initialize()
	for i = 0, Map.GetNumPlots()-1, 1 do
		local plot = Map.GetPlotByIndex(i);
		fogInstances[i] = fogManager:GetInstance()
		hexInstances[i] = hexManager:GetInstance()
		riverInstances[i] = riverManager:GetInstance()

		local wx, wy, wz = GridToWorld(plot:GetX(), plot:GetY());
		fogInstances[i].Anchor:SetWorldPositionVal(wx, wy, 0);
		hexInstances[i].Anchor:SetWorldPositionVal(wx, wy, 0);
		riverInstances[i].Anchor:SetWorldPositionVal(wx, wy, 0);

		for j = 0, 5, 1 do
			local ctl = riverInstances[i]["Vertex"..j];
			ctl:SetOffsetVal(riverImagesLocations[j].x, riverImagesLocations[j].y);
			ctl:SetSizeVal(238, 238);
		end

		hexInstances[i].Civ:SetSizeVal(400, 400);
		hexInstances[i].Type:SetSizeVal(425, 425);
		hexInstances[i].Feature:SetSizeVal(425, 425);
		hexInstances[i].Resource:SetSizeVal(148, 148);
		hexInstances[i].Resource:SetTextureOffsetVal(40, 40);
		hexInstances[i].Improvement:SetSizeVal(148, 148);
		hexInstances[i].Improvement:SetTextureOffsetVal(40, 40);

		fogInstances[i].Fog:SetSizeVal(500, 500);
	end
end

-------------------------------------------------------------------------------------------------
local function Update()
	if ContextPtr:IsHidden() then return end
	UpdateResourceVisibility();

	startingPlots = {};
	for i = 0, GameDefines.MAX_CIV_PLAYERS, 1 do
		local player = Players[i];
		if player:IsEverAlive() then
			if startingPlotGetter(player) then
				startingPlots[startingPlotGetter(player)] = player;
			end
		end
	end

	for i = 0, Map.GetNumPlots()-1, 1 do
		UpdatePlot(Map.GetPlotByIndex(i), hexInstances[i], riverInstances[i], fogInstances[i])
	end
end
LuaEvents.RSD_UpdateResources.Add(Update)

-------------------------------------------------------------------------------------------------
local TOP    = 0;
local BOTTOM = 1;
local LEFT   = 2;
local RIGHT  = 3;

local function ScrollMouseEnter(which)
    if which == TOP then
		Events.SerialEventCameraStartMovingForward();
    elseif which == BOTTOM then
		Events.SerialEventCameraStartMovingBack();
    elseif which == LEFT then
		Events.SerialEventCameraStartMovingLeft();
    else
		Events.SerialEventCameraStartMovingRight();
    end
end

local function ScrollMouseExit(which)
    if which == TOP then
		Events.SerialEventCameraStopMovingForward();
    elseif which == BOTTOM then
		Events.SerialEventCameraStopMovingBack();
    elseif which == LEFT then
		Events.SerialEventCameraStopMovingLeft();
    else
		Events.SerialEventCameraStopMovingRight();
    end
end
Controls.ScrollTop:RegisterCallback( Mouse.eMouseEnter, ScrollMouseEnter);
Controls.ScrollTop:RegisterCallback( Mouse.eMouseExit, ScrollMouseExit);
Controls.ScrollTop:SetVoid1(TOP);

Controls.ScrollBottom:RegisterCallback( Mouse.eMouseEnter, ScrollMouseEnter);
Controls.ScrollBottom:RegisterCallback( Mouse.eMouseExit, ScrollMouseExit);
Controls.ScrollBottom:SetVoid1(BOTTOM);

Controls.ScrollLeft:RegisterCallback( Mouse.eMouseEnter, ScrollMouseEnter);
Controls.ScrollLeft:RegisterCallback( Mouse.eMouseExit, ScrollMouseExit);
Controls.ScrollLeft:SetVoid1(LEFT);

Controls.ScrollRight:RegisterCallback( Mouse.eMouseEnter, ScrollMouseEnter);
Controls.ScrollRight:RegisterCallback( Mouse.eMouseExit, ScrollMouseExit);
Controls.ScrollRight:SetVoid1(RIGHT);

-------------------------------------------------------------------------------------------------
local initialized = false
local function OnShowMap(_map, _startingPlotGetter)
	print("RESEED: StrategicMap.OnShowPreview")
	Map = _map
	RsdMap = _map
	startingPlotGetter = _startingPlotGetter

	if not initialized then
		initialized = true
		Initialize() 
	end

	ContextPtr:SetHide(false)
	print("RESEED: Updating strategic view")
	Update()
	print("RESEED: Updated strategic view")
end
LuaEvents.RSD_ShowPreview.Add(OnShowMap)

-------------------------------------------------------------------------------------------------
local function OnHideMap()
	ContextPtr:SetHide(true)
end
LuaEvents.RSD_HidePreview.Add(OnHideMap)

-------------------------------------------------------------------------------------------------
local function OnRevealMapChanged(_revealMap)
	revealMap = _revealMap
	Update()
end
LuaEvents.RSD_RevealMapChanged.Add(OnRevealMapChanged)