-- Released under GPL v3
--------------------------------------------------------------
function L(str, ...)
	if str == nil then return "???" end
	if type(str) ~= "string" then return tostring(str) end
	return Locale.ConvertTextKey(str, ...);
end

-------------------------------------------------------------------------------------------------
function dump(x)
	if x == nil then 
		print("nil");
	elseif type(x) == "table" then
		for k, v in pairs(x) do
			print(k.." = "..tostring(v));
		end
	else
		print(tostring(x));
	end
end

-------------------------------------------------------------------------------------------------
function dumpMembers(x)
	for k in pairs(x) do
		dump(k);
	end
end

-------------------------------------------------------------------------------------------------
function dumpMetaIndex(x)
	for k in pairs(getmetatable(x).__index) do
		dump(k);
	end
end

-------------------------------------------------------------------------------------------------
function LoadFile(fileName)
	local files = {};
	UI.SaveFileList(files, GameTypes.GAME_SINGLE_PLAYER, false, true);

	for _, path in ipairs(files) do
		local thisFileName = Path.GetFileNameWithoutExtension(path);
		if thisFileName == fileName then
			print("Loading "..path);
			Events.PlayerChoseToLoadGame(path);
			return path;
		end
	end
end

-------------------------------------------------------------------------------------------------
function DeleteFiles(fileName)
	print("Deleting");
	assert(false)
	local files = {};
	UI.SaveFileList(files, GameTypes.GAME_SINGLE_PLAYER, false, true);

	for _, path in ipairs(files) do
		local otherName = Path.GetFileNameWithoutExtension(path);
		if Locale.ToUpper(otherName) == Locale.ToUpper(fileName) then
			print("Deleting "..otherName);
			print(path);
			UI.DeleteSavedGame(path);
		end
	end
end

-------------------------------------------------------------------------------------------------
function GetFullPath(fileName)
	local files = {};
	UI.SaveFileList(files, GameTypes.GAME_SINGLE_PLAYER, false, true);

	for _, path in ipairs(files) do
		local otherName = Path.GetFileNameWithoutExtension(path);
		if Locale.ToUpper(otherName) == Locale.ToUpper(fileName) then
			return path
		end
	end
end

-------------------------------------------------------------------------------------------------
function lastIndexOf(str, c)
	str = string.reverse(str);
	local index = string.find(str, c);
	if not index then return 0 end
	return string.len(str) + 1 - index;
end

-------------------------------------------------------------------------------------------------
function getStackTrace(thread)
	local trace = {};
	local level = 1
	while true do
		local info = thread and debug.getinfo(thread, level, "nSl") or debug.getinfo(level, "nSl")
		if not info then break end
		if info.what == "C" then   -- is a C function?
			table.insert(trace, "C function");
		else   
			-- a Lua function
			local userStr = Path.GetFileName(info.source)..": "..info.currentline;
			if (info.name and string.len(info.name)) then
				userStr = userStr.." ("..info.name..")";
			end
			table.insert(trace, userStr);
	    end
		level = level + 1
	end
	return trace;
end

-------------------------------------------------------------------------------------------------
function FormatError(err, levelsToIgnoreBelow, thread)
	levelsToIgnoreBelow = levelsToIgnoreBelow or 0
	print(err);

	-- Slice error to make it change its formatting later
	local index = string.find(err, "...", 1, true);				-- Path truncated with '...'
	if not index then index = -5 end
	local index2 = index + 6;									-- Line error starts here
	local index3 = string.find(err, ":", index2, true);			-- Error message starts here
	if index3 then 
		index3 = index3 + 1 
	else
		index3 = 1
	end

	-- Fix error message, prints it, appends the trace when available
	if debug and debug.getinfo then
		err = string.sub(err, index3);

		-- Stack trace: 1 is this function, 2 is the error location, 3 is the C error handler, 4 and greater are LUA calls
		local trace = getStackTrace(thread);
		for i, v in ipairs(trace) do
			print(v);
			if (i > levelsToIgnoreBelow) then
				err = err.."[NEWLINE]"..v;
			end
		end
	else
		err = "line "..string.sub(err, index2);
    end

	return err;
end

-------------------------------------------------------------------------------------------------
function Shuffle(t)
	local len = #t;
	local random = Map.Rand;
	for i = 1, len, 1 do
		local k = random(len - 1, "Shuffling Values") + 1;
		t[i], t[k] = t[k], t[i];
	end
end

-------------------------------------------------------------------------------------------------
function NextReversed(t, lastVal, lastKey)
	local k2, v2 = next(t, lastKey)
	return t, v2, k2
end


-------------------------------------------------------------------------------------------------
local completeKillsID = GameInfo.GameOptions["GAMEOPTION_COMPLETE_KILLS"].ID;
local dominationVictoryID = GameInfo.Victories["VICTORY_DOMINATION"].ID;

function CanDeleteCities()
	return Game.IsOption(completeKillsID) and (not PreGame.IsVictory(dominationVictoryID))
end

function RelaxOptions()
	Game.SetOption(completeKillsID, true);
	PreGame.SetVictory(dominationVictoryID, false);
end
